"""Bluetooth Low Energy Secure Manager exceptions
"""

class SMInvalidParameterFormat(Exception):
    def __init__(self):
        super().__init__()

class SMInvalidCustomFunction(Exception):
    def __init__(self):
        super().__init__()

__all__ = [
    'SMInvalidParameterFormat',
    'SMInvalidCustomFunction'
]



# Sorting logic auto-generated via AI, and results just eyeballed to see that they look good enough. Most likely has issues
# The basic goal is to sort first by company name, alphabetically, ascending order
# Then second to group things so that UUIDs with parent_UUID fields are placed underneath the parent_UUIDs
# and the placed-under elements should be sorted also in ascending order.

import json
from itertools import groupby
from collections import defaultdict

def merge_entries(entry1, entry2):
    # Check if UUID_usage_array fields are the same
    if entry1.get('UUID_usage_array') != entry2.get('UUID_usage_array'):
        return entry1  # Skip merge if UUID_usage_array fields are not the same

    merged_entry = {}
    for key in entry1:
        if isinstance(entry1[key], list) and isinstance(entry2[key], list):
            merged_entry[key] = list({json.dumps(item) for item in entry1[key] + entry2[key]})
            merged_entry[key] = [json.loads(item) for item in merged_entry[key]]
        elif isinstance(entry1[key], str) and isinstance(entry2[key], str):
            if entry1[key] == entry2[key]:
                merged_entry[key] = entry1[key]
            else:
                merged_entry[key] = f"MERGED: ({entry1[key]}), ({entry2[key]})"
        else:
            merged_entry[key] = entry1[key] if entry1[key] is not None else entry2[key]
    return merged_entry

def entries_are_equal(entry1, entry2):
    if not entry1 or not entry2:
        return False
    return all(entry1.get(key) == entry2.get(key) for key in entry1)

def sort_custom_uuids(file_path):
    with open(file_path, 'r') as f:
        data = json.load(f)

    def sort_key(entry):
        company = entry['company'] if entry['company'] is not None else ''
        return company, entry['UUID_purpose'], entry['UUID']

    # Convert UUID and parent_UUID to lowercase and group entries by concatenated key
    uuid_groups = {}
    for entry in data:
        entry['UUID'] = entry['UUID'].lower()
        if 'parent_UUID' in entry:
            entry['parent_UUID'] = entry['parent_UUID'].lower()
        company = entry['company'] if entry['company'] is not None else ''
        uuid_usage_array = json.dumps(entry.get('UUID_usage_array', []))
        key = f"{company}-{entry['UUID']}-{uuid_usage_array}"
        if key in uuid_groups:
            uuid_groups[key] = merge_entries(uuid_groups[key], entry)
        else:
            uuid_groups[key] = entry

    # Sort the entries
    sorted_entries = sorted(uuid_groups.values(), key=sort_key)

    # Group entries by parent_UUID
    parent_uuid_map = defaultdict(list)
    for entry in sorted_entries:
        parent_uuid = entry.get('parent_UUID')
        if parent_uuid:
            parent_uuid_map[parent_uuid].append(entry)
        else:
            parent_uuid_map[entry['UUID']].append(entry)

    # Flatten the grouped entries
    final_sorted_entries = []
    for uuid, entries in parent_uuid_map.items():
        # Add the parent entry first
        parent_entry = next((e for e in entries if e['UUID'] == uuid), None)
        if parent_entry:
            final_sorted_entries.append(parent_entry)
            entries.remove(parent_entry)
        # Sort the child entries by UUID
        sorted_children = sorted(entries, key=lambda e: e['UUID'])
        # Add the child entries
        final_sorted_entries.extend(sorted_children)

    # Write the sorted entries back to the file
    with open(file_path, 'w') as f:
        json.dump(final_sorted_entries, f, indent=4)

# Example usage
# sort_custom_uuids('/path/to/your/file.json')

if __name__ == "__main__":
    sort_custom_uuids('CLUES_data.json')



"""Dot15d4 sniffing module

This module provides a Sniffer class for Dot15d4 protocol, that is used by
`whad-sniff` to sniff dot15d4-based protocols.
"""
from time import time
from typing import Generator, List

# Required by type hints
from scapy.packet import Packet

from whad.dot15d4.connector import Dot15d4
from whad.dot15d4.sniffing import SnifferConfiguration
from whad.exceptions import UnsupportedCapability, WhadDeviceDisconnected
from whad.helpers import message_filter
from whad.common.sniffing import EventsManager
from whad.hub.dot15d4 import RawPduReceived, PduReceived


class Sniffer(Dot15d4, EventsManager):
    """
    802.15.4 Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        """Initialize our sniffer.
        """
        Dot15d4.__init__(self, device)
        EventsManager.__init__(self)

        self.__configuration = SnifferConfiguration()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        self.sniff_dot15d4(channel=self.__configuration.channel)

    @property
    def configuration(self) -> SnifferConfiguration:
        """Retrieve this sniffer configuration.
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: SnifferConfiguration):
        """Set the sniffer configuration.
        """
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    @property
    def channel(self) -> int:
        """Retrieve the currently configured channel this sniffer is listening on.
        """
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel: int = 11):
        """Set current channel and start sniffing on it.
        """
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    def available_actions(self, filter=None) -> List:
        """Retrieve the possible actions on this sniffer.
        """
        actions = []
        return [action for action in actions if filter is None or isinstance(action, filter)]

    def sniff(self, timeout: float = None) -> Generator[Packet, None , None]:
        """Main sniffing loop.

        This method waits for raw PDUs or PDUs, depending on hardware capability, and report
        them to the caller by yielding Scapy packets created from sniffed PDUs. These packets
        also include metadata (reception timestamp, channel, ...).

        :param timeout: Timeout in seconds (disabled if set to `None`)
        :type timeout: float, optional
        """
        start = time()
        try:
            while True:
                if self.support_raw_pdu():
                    message_type = RawPduReceived
                else:
                    message_type = PduReceived

                # Wait for a message to be received
                message = self.wait_for_message(keep=message_filter(message_type), timeout=1)
                if message is not None:
                    packet = message.to_packet()
                    if packet is not None:
                        self.monitor_packet_rx(packet)
                        yield packet

                # Check if timeout has been reached (if provided)
                if timeout is not None:
                    if time() - start >= timeout:
                        break
        except WhadDeviceDisconnected:
            return



"""
Bluetooth Low Energy Tiny Stack for Unit Testing
================================================

GATT client has no profile but implements a basic
discovery procedure that will populate its internal
attribute database.
"""

import logging
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import ATT_Hdr

from whad.ble.profile.attribute import UUID

from .attribute import Attribute

# ATT Procedures
from .procedure import UnexpectedProcError
from .read_by_group_type import ClientReadByGroupTypeProcedure
from .read_by_type import ClientReadByTypeProcedure
from .readblob import ClientReadBlobProcedure
from .read import ClientReadProcedure
from .write import ClientWriteProcedure
from .writecmd import ClientWriteCommandProcedure
from .find_info import ClientFindInformationProcedure, ClientFindByTypeValueProcedure
from .notif import ClientNotificationCheckProcedure, ClientIndicationCheckProcedure

logger = logging.getLogger(__name__)

class GattClient:
    """Tiny GATT client"""

    def __init__(self, l2cap: 'Llcap' = None):
        """Initialize a GATT client."""

        self.__l2cap = l2cap

        # Initialize client state
        self.__attributes = []
        self.__mtu = 23
        self.__cur_procedure = None

        # Register procedures
        self.__procedures = [
            ClientReadByGroupTypeProcedure,
            ClientReadByTypeProcedure,
            ClientReadProcedure,
            ClientReadBlobProcedure,
            ClientWriteProcedure,
            ClientWriteCommandProcedure,
            ClientFindInformationProcedure,
            ClientFindByTypeValueProcedure,
            ClientNotificationCheckProcedure,
            ClientIndicationCheckProcedure,
        ]

    @property
    def attributes(self) -> List[Attribute]:
        """Remote server attributes."""
        return self.__attributes

    def set_l2cap(self, obj):
        """Set L2CAP layer for this GattClient."""
        self.__l2cap = obj

    def on_pdu(self, request: Packet) -> List[Packet]:
        """Process incoming response/error."""
        # If a suitable procedure has been found or is in progress, forward request.
        if self.__cur_procedure is not None:
            # Forward to current procedure
            answers: list[Packet] = self.__cur_procedure.process_request(request)

            # Unset current procedure if finished
            if self.__cur_procedure.done():
                #self.__cur_procedure = None
                """nope"""

            # Automatically add ATT_Hdr()
            if isinstance(answers, list):
                return [ ATT_Hdr()/ans if ATT_Hdr not in ans else ans for ans in answers]
            else:
                # Should not happen so raise an exception
                raise UnexpectedProcError()
        else:
            logger.warning("[ble::mock::stack::gatt_client] No procedure to handle packet ! (%s)", request)
            return []

    def read_by_group_type(self, group_type: UUID, start_handle: int, end_handle: int) -> List[Packet]:
        """Enumerate attributes by group type."""
        # Initiate a ClientReadByGroupTypeProcedure
        self.__cur_procedure = ClientReadByGroupTypeProcedure(
            group_type, start_handle,end_handle
        )

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)


    def read_by_type(self, start_handle: int, end_handle: int, type_uuid: UUID) -> List[Packet]:
        """Read attributes by type."""
        # Initiate a ClientReadByTypeProcedure
        self.__cur_procedure = ClientReadByTypeProcedure(
            start_handle, end_handle, type_uuid
        )

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def read(self, handle: int) -> List[Packet]:
        """Read attribute."""
        # Initiate a ClientReadProcedure
        self.__cur_procedure = ClientReadProcedure(handle)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def read_blob(self, handle: int, offset: int) -> List[Packet]:
        """Read part of attribute."""
        # Initiate a ClientReadBlobProcedure
        self.__cur_procedure = ClientReadBlobProcedure(handle, offset)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def find_information(self, start_handle: int, end_handle: int):
        """Find information (attribute type) about a list of handles.

        :param start_handle: Start handle
        :type start_handle: int
        :param end_handle: End handle, must be lower or equal to start handle
        :type end_handle: int
        :return: List of packets to send once the procedure initiated
        :rtype: list
        """
        # Initiate a ClientFindInformationProcedure
        self.__cur_procedure = ClientFindInformationProcedure(start_handle, end_handle)

        # Generate ATT packets to send when this procedure is initiated and forward
        # them to the underlying L2CAP layer.
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def find_by_type_value(self, start_handle: int, end_handle: int, attr_type: UUID, attr_value: bytes):
        """Find attribute by type UUID.

        :param start_handle: Start handle
        :type start_handle: int
        :param end_handle: End handle, must be lower or equal to start handle
        :type end_handle: int
        :param attr_type: Attribute Type UUID
        :type attr_type: UUID
        :return: List of packets to send once the procedure initiated
        :rtype: list
        """
        # Initiate a ClientFindInformationProcedure
        self.__cur_procedure = ClientFindByTypeValueProcedure(start_handle, end_handle, attr_type, attr_value)

        # Generate ATT packets to send when this procedure is initiated and forward
        # them to the underlying L2CAP layer.
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)


    def wait_procedure(self, timeout: float = None):
        """Wait for the current procedure to complete."""
        return self.__cur_procedure.wait(timeout=timeout)

    def write(self, handle: int, value: bytes) -> List[Packet]:
        """Write value into a specified attribute."""
        # Initiate a ClientWriteProcedure
        self.__cur_procedure = ClientWriteProcedure(handle, value)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def write_cmd(self, handle: int, value: bytes) -> List[Packet]:
        """Write value into a specific attribute without waiting
        for a response.

        :param handle: Handle of the target attribute
        :type handle: int
        :param value: Value to write into the target attribute
        :type value: bytes
        """
        # Initiate a ClientWriteCommandProcedure
        self.__cur_procedure = ClientWriteCommandProcedure(handle, value)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def sub_notif(self, handle: int):
        """Subscribe to notification for the specified handler."""
        # Initiate a ClientNotificationCheckProcedure
        self.__cur_procedure = ClientNotificationCheckProcedure(handle)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)

    def sub_ind(self, handle: int):
        """Subscribe to notification for the specified handler."""
        # Initiate a ClientNotificationCheckProcedure
        self.__cur_procedure = ClientIndicationCheckProcedure(handle)

        # Generate ATT packets to send when this procedure is initiated
        # and forward them to the underlying link-layer
        for req in self.__cur_procedure.initiate():
            self.__l2cap.send_pdu(ATT_Hdr()/req)




"""
This module provides a single :py:class:`~whad.phy.connector.sniffer.Sniffer`
class that supports the following modulations:

- ASK
- GFSK
- BFSK
- QFSK
- BPSK
- QPSK
- LoRa

This sniffer is configured using a dedicated configuration class,
:py:class:`~whad.phy.sniffing.SnifferConfiguration`. This class allows users
to configure some default fields like the frequency, data rate and synchronization
word, but also some packet-specific parameters like the maximum packet size or
its endianness:

>>> config = SnifferConfiguration()
>>> config.frequency = 2402000000
>>> config.datarate = 1000000
>>> config.endianness = Endianness.LITTLE
>>> config.packet_size = 200
>>> config.sync_word = b'\\xAA\\xAA'

Modulation is selected by setting the correct modulation scheme to `True`:

>>> config.gfsk = True

Some modulations offer extra parameters, like the FSK and LoRa modulations.
FSK modulation can be customized through a :py:class:`~whad.phy.sniffing.FSKConfiguration`
object and LoRa through a :py:class:`~whad.phy.sniffing.LoRaConfiguration` object.

Once the sniffer configuration set, sniffing is quite easy:

>>> sniffer = Sniffer(WhadDevice.create("uart0"))
>>> sniffer.configuration = config
>>> sniffer.start()
>>> for packet in sniffer.sniff():
        packet.show()
"""
import logging

from time import time
from queue import Queue, Empty
from typing import Generator
from scapy.packet import Packet

from whad.exceptions import WhadDeviceDisconnected
from whad.hub.phy import Endianness
from whad.common.sniffing import EventsManager
from whad.exceptions import UnsupportedCapability

from .base import Phy
from ..sniffing import SnifferConfiguration

logger = logging.getLogger(__name__)

class Sniffer(Phy, EventsManager):
    """
    Phy Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        Phy.__init__(self, device)
        EventsManager.__init__(self)

        # Queue to hold sniffed packets
        self.__packets = Queue()

        # Default configuration
        self.__configuration = SnifferConfiguration()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def __enable_sniffing(self):
        """Enable sniffing.

        Configure the associated hardware and set the sniffer parameters.
        """
        # Set frequency (will raise UnsupportedFrequency if provided frequency is outside
        # hardware's supported ranges).
        self.set_frequency(self.__configuration.frequency)

        # Set packet size, display a warning if hardware rejects the provided value.
        if not self.set_packet_size(self.__configuration.packet_size):
            logger.warning("[WARNING] Hardware rejected the provided packet size (%d) and will use its maximum packet size instead.",
                            self.__configuration.packet_size)

        # Set data rate for all modulations but LoRa.
        if not self.__configuration.lora:
            self.set_datarate(self.__configuration.datarate)

        # Configure modulation.
        if self.__configuration.gfsk:
            self.set_gfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.bfsk:
            self.set_bfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.qfsk:
            self.set_4fsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.ask:
            self.set_ask()
        elif self.__configuration.bpsk:
            self.set_bpsk()
        elif self.__configuration.qpsk:
            self.set_qpsk()
        elif self.__configuration.lora:
            self.set_lora(
                self.__configuration.lora_configuration.spreading_factor,
                self.__configuration.lora_configuration.coding_rate,
                self.__configuration.lora_configuration.bandwidth,
                self.__configuration.lora_configuration.preamble_length,
                crc=self.__configuration.lora_configuration.enable_crc,
                explicit=self.__configuration.lora_configuration.enable_explicit_mode
            )

        # Set endianness for all modulations but LoRa.
        if not self.__configuration.lora:
            self.set_endianness(
                Endianness.LITTLE if
                self.__configuration.little_endian else
                Endianness.BIG
            )

        # Set synchronization word.
        self.set_sync_word(self.__configuration.sync_word)

        # Put WHAD interface in sniffing mode.
        self.sniff_phy()

    @property
    def configuration(self):
        """Current sniffing configuration
        """
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration: SnifferConfiguration):
        """Set the sniffer's configuration.

        It will stop the sniffer if it is enabled, update its configuration
        and resume sniffing. Sniffing starts automatically each time a new
        configuration is set.
        """
        self.stop()
        self.__configuration = new_configuration
        self.__enable_sniffing()

    @property
    def frequency(self) -> int:
        """Configured frequency in Hertz.
        """
        return self.__configuration.frequency

    @frequency.setter
    def frequency(self, frequency: int = 2402000000):
        """Update snffer's frequency. If already active, sniffing is stopped and resumed.
        """
        self.stop()
        self.__configuration.frequency = frequency
        self.__enable_sniffing()

    def available_actions(self, action_filter=None) -> list:
        """List available actions.
        """
        actions = []
        return [action for action in actions if action_filter is None or isinstance(action, filter)]

    def on_packet(self, packet: Packet):
        """Packet reception handler: put packets in sniffing queue.

        :param  packet: Received packet
        :type   packet: Packet
        """
        self.__packets.put(packet)

    def sniff(self, timeout: float = None) -> Generator[Packet, None, None]:
        """Sniff packets out of thin air.

        :param timeout: Specify the number of seconds after which sniffing will stop.
                        Wait forever if set to `None`.
        :type timeout: float
        """
        try:
            while (timeout is None) or (timeout>0.0):
                # Wait for a packet
                start = time()
                packet = self.__packets.get(True, timeout=timeout)

                # Update remaining time if required
                if timeout is not None:
                    timeout = timeout - (time() - start)

                # Notify packet reception
                yield packet
        except Empty:
            # Timeout reached, exit
            return
        except WhadDeviceDisconnected:
            # Device disconnected, exit
            return



from whad.zigbee.exceptions import MissingNetworkSecurityHeader
from Cryptodome.Cipher import AES
from scapy.layers.dot15d4 import Dot15d4,Dot15d4FCS
from scapy.layers.zigbee import ZigbeeSecurityHeader,ZigbeeNWK, ZigbeeAppCommandPayload, \
    ZigbeeAppDataPayload, ZigbeeNWKCommandPayload
from whad.scapy.layers.zll import ZLLScanRequest, ZLLScanResponse, ZLLNetworkJoinRouterRequest
from whad.common.analyzer import TrafficAnalyzer
from scapy.compat import raw
from scapy.config import conf
from scapy.all import Packet
from struct import pack

conf.dot15d4_protocol = "zigbee"

def e(key, plaintext):
    cipher = AES.new(plaintext, AES.MODE_ECB)
    output = cipher.encrypt(key)
    return output

def hash(input):
    output = b"\x00"*16
    M = input
    M += b"\x80"
    while len(M) % 16 != 0:
        M += b"\x00"
    M = M[:-2] + bytes([(0xFF & (len(input)*8 >> 8)),(0xFF & (len(input)*8 >> 0))])
    i = 0
    while i < len(M):
        bloc = M[i:i+16]
        ciphertext = e(bloc,output)
        output = b""
        for j in range(len(bloc)):
            output += bytes([ciphertext[j] ^ bloc[j]])
        i+=16
    return output

def hash_key(key, input):
    ipad = 0x36
    opad = 0x5c

    hash_in = hash_out = b""
    for i in key:
        hash_in += bytes([opad ^ i])
        hash_out += bytes([ipad ^ i])

    hash_out += bytes([input])
    hash_in += hash(hash_out)
    return hash(hash_in)


class CryptoManager:
    SECURITY_LEVELS = {
    	0x00: {"encryption":False,"integrity":False, "M":0},
    	0x01: {"encryption":False,"integrity":True, "M":4},
    	0x02: {"encryption":False,"integrity":True, "M":8},
    	0x03: {"encryption":False,"integrity":True, "M":16},
    	0x04: {"encryption":True,"integrity":False, "M":0},
    	0x05: {"encryption":True,"integrity":True, "M":4},
    	0x06: {"encryption":True,"integrity":True, "M":8},
    	0x07: {"encryption":True,"integrity":True, "M":16}
    }


    def __init__(self, key):
        self.key = key
        self.base_class = None
        self.nonce = None
        self.auth = None
        self.M = None
        self.encryption = None
        self.integrity = None
        self.patched = False

    def generateNonce(self, packet):
        source = raw(packet[ZigbeeSecurityHeader])[5:13] # prevent struct failure on windows ...
        fc = pack("I", packet[ZigbeeSecurityHeader].fc)
        security_header = bytes([(
            packet[ZigbeeSecurityHeader].nwk_seclevel |
            (packet[ZigbeeSecurityHeader].key_type << 3) |
            (packet[ZigbeeSecurityHeader].extended_nonce << 5)|
            (packet[ZigbeeSecurityHeader].reserved1 << 6)
        )])
        return source + fc + security_header

    def checkSecurityLevel(self, packet):
        # If network security level equals to zero, we assume it's the bug mentioned in wireshark source code and patch it
        if packet[ZigbeeSecurityHeader].nwk_seclevel == 0:
            self.patched = True
            packet[ZigbeeSecurityHeader].nwk_seclevel = 0x05 # patched security header: MIC-ENC-32
            M = 4
            encryption = True
            integrity = True
        else:
            # Parse network security level to check how to process the packet
            self.patched = False
            level = CryptoManager.SECURITY_LEVELS[packet[ZigbeeSecurityHeader].nwk_seclevel]
            encryption = level["encryption"]
            integrity = level["integrity"]
            if integrity:
                M = level["M"]
            else:
                M = 0
        return (packet, M, integrity, encryption)

    def generateAuth(self, packet):
        if self.encryption:
            auth = raw(packet[self.base_class:]).replace(packet.data,b"")
            if len(packet.mic) != 0:
                auth = auth.replace(packet.mic, b"")
        else:
            auth = raw(packet[self.base_class:])[:-self.M]
        return auth

    def extractCiphertextPayload(self, packet):
        if self.encryption:
            message = packet[ZigbeeSecurityHeader].data[:-self.M] if len(packet.mic) == 0 and self.patched else packet.data
            mic = packet[ZigbeeSecurityHeader].data[-self.M:] if len(packet.mic) == 0 and self.patched else packet.mic
        else:
            mic = raw(packet[self.base_class:])[-self.M:]
            message = b""
        return message, mic

    def generateMIC(self,packet):
        self.auth = self.generateAuth(packet)
        plaintext = packet.data if len(packet.mic) == 0 and self.patched else packet.data + packet.mic
        auth = len(self.auth).to_bytes(2, byteorder = 'big')+self.auth
        auth = auth + (32 - len(auth))*b"\x00" + plaintext+(16 - len(plaintext))*b"\x00"
        flags = (0 << 7) | ((0 if len(self.auth) == 0 else 1) << 6 ) | ((2 - 1) << 3) | ((self.M-2)//2 if self.M else 0)
        B0 = bytes([flags]) + self.nonce + pack(">H",len(plaintext))
        X0 = b"\x00"*16
        cipher = AES.new(self.key, AES.MODE_CBC, X0)
        padding = b""
        while len(padding + B0 + auth) % 16 != 0:
            padding += b"\x00"
        X1 = cipher.encrypt(padding + B0 + auth)
        return X1[-16:-12]

    def encrypt(self, packet):
        # convert into scapy packet if bytes only
        if isinstance(packet, bytes):
            packet = Dot15d4(packet)

        # don't process FCS if present
        if Dot15d4FCS in packet:
            packet = Dot15d4(raw(packet)[:-2])

        # raise MissingNetworkSecurityHeader exception if no security header is found
        if ZigbeeSecurityHeader not in packet:
            raise MissingNetworkSecurityHeader()

        if self.base_class is None:
            self.base_class = packet[ZigbeeSecurityHeader].underlayer.__class__

        # check security level (and patch the packet content if needed - see wireshark source code for details)
        packet, self.M, self.integrity, self.encryption = self.checkSecurityLevel(packet)

        # generate the nonce
        self.nonce = self.generateNonce(packet)

        # generate the AES-CCM parameters
        self.auth = self.generateAuth(packet)

        # Extract plaintext
        plaintext = packet.data[:-self.M] if len(packet.mic) == 0 and self.patched else packet.data
        # Encrypt and generate MIC
        cipher = AES.new(self.key, AES.MODE_CCM, nonce=self.nonce, mac_len=self.M)
        cipher.update(self.auth)

        ciphertext = cipher.encrypt(plaintext)

        mic = cipher.digest()
        packet.data = ciphertext
        packet.mic = mic

        # Restore security level if patched
        if self.patched:
            packet[ZigbeeSecurityHeader].nwk_seclevel = 0

        return packet

    def decrypt(self, packet):
        # convert into scapy packet if bytes only
        if isinstance(packet, bytes):
            packet = Dot15d4(packet)

        # don't process FCS if present
        if Dot15d4FCS in packet:
            packet = Dot15d4(raw(packet)[:-2])

        # raise MissingNetworkSecurityHeader exception if no security header is found
        if ZigbeeSecurityHeader not in packet:
            raise MissingNetworkSecurityHeader()

        if self.base_class is None:
            self.base_class = packet[ZigbeeSecurityHeader].underlayer.__class__

        # check security level (and patch the packet content if needed - see wireshark source code for details)
        packet, self.M, self.integrity, self.encryption = self.checkSecurityLevel(packet)

        # generate the nonce
        self.nonce = self.generateNonce(packet)

        # generate the AES-CCM parameters
        self.auth = self.generateAuth(packet)
        ciphertext, mic = self.extractCiphertextPayload(packet)

        # Perform the decryption and integrity check
        cipher = AES.new(self.key, AES.MODE_CCM, nonce=self.nonce, mac_len=self.M)
        cipher.update(self.auth)

        plaintext = cipher.decrypt(ciphertext)
        try:
            cipher.verify(mic)
            packet.data = plaintext
            packet.mic = self.generateMIC(packet)
            # Reverse patching if needed
            if self.patched:
                packet[ZigbeeSecurityHeader].nwk_seclevel = 0

            return (packet, True)

        except ValueError:
            # Reverse patching if needed
            if self.patched:
                packet[ZigbeeSecurityHeader].nwk_seclevel = 0
            return (packet, False) # integrity check

class NetworkLayerCryptoManager(CryptoManager):

    def __init__(self, key):
        super().__init__(key)
        self.base_class = ZigbeeNWK

class ApplicationSubLayerCryptoManager(CryptoManager):

    def __init__(self, key, input):
        if input is None:
            generated_key = self.base_key = key
        else:
            self.input = input
            self.base_key = key
            generated_key = hash_key(key, self.input)
        super().__init__(generated_key)
        self.base_class = ZigbeeAppDataPayload

class ZigbeeDecryptor:
    def __init__(self, *keys):
        self.keys = list(keys)

    def add_key(self, key):
        if isinstance(key, str):
            if len(key) == 16:
                key = key.encode('ascii')
            else:
                try:
                    key = bytes.fromhex(key.replace(":",""))
                except ValueError:
                    return False

        if not isinstance(key, bytes) or len(key) != 16:
            return False

        if key not in self.keys:
            self.keys.append(key)
            return True
        return False

    def attempt_to_decrypt(self, packet):

        if ZigbeeSecurityHeader not in packet:
            raise MissingNetworkSecurityHeader()

        if packet[ZigbeeSecurityHeader].underlayer.__class__ == ZigbeeNWK:
            for key in self.keys:
                manager = NetworkLayerCryptoManager(key)
                decrypted, success = manager.decrypt(packet)
                if success:
                    if packet.frametype == 0:
                        return ZigbeeAppDataPayload(decrypted.data), True
                    elif packet.frametype == 1:
                        return ZigbeeNWKCommandPayload(decrypted.data), True
                    else:
                        return decrypted.data, True
        else:
            for key in self.keys:
                manager = ApplicationSubLayerCryptoManager(key, 1)
                decrypted, success = manager.decrypt(packet)
                if success:
                    if packet.frametype == 0:
                        if packet.aps_frametype == 0:
                            return ZigbeeAppDataPayload(decrypted.data), True
                        else:
                            return ZigbeeAppCommandPayload(decrypted.data), True
                    elif packet.frametype == 1:
                        return ZigbeeNWKCommandPayload(decrypted.data), True
                    else:
                        return decrypted.data, True

                manager = ApplicationSubLayerCryptoManager(key, 0)
                decrypted, success = manager.decrypt(packet)
                if success:
                    if packet.frametype == 0:
                        if packet.aps_frametype == 0:
                            return ZigbeeAppDataPayload(decrypted.data), True
                        else:
                            return ZigbeeAppCommandPayload(decrypted.data), True
                    elif packet.frametype == 1:
                        return ZigbeeNWKCommandPayload(decrypted.data), True
                    else:
                        return decrypted.data, True

                manager = ApplicationSubLayerCryptoManager(key, 2)
                decrypted, success = manager.decrypt(packet)
                if success:
                    if packet.frametype == 0:
                        if packet.aps_frametype == 0:
                            return ZigbeeAppDataPayload(decrypted.data), True
                        else:
                            return ZigbeeAppCommandPayload(decrypted.data), True
                    elif packet.frametype == 1:
                        return ZigbeeNWKCommandPayload(decrypted.data), True
                    else:
                        return decrypted.data, True

        return packet, False


ZIGBEE_ZLL_KEYS = {
    "master_key":           b"\x9F\x55\x95\xF1\x02\x57\xC8\xA4\x69\xCB\xF4\x2B\xC9\x3F\xEE\x31",
    "certification_key":    b"\xC0\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xCA\xCB\xCC\xCD\xCE\xCF",
}


class TransportKeyDistribution(TrafficAnalyzer):
    def reset(self):
        super().reset()
        self.transport_key = None

    def process_packet(self, packet):

        if ZigbeeSecurityHeader in packet and packet[ZigbeeSecurityHeader].key_type == 2:
            self.trigger()
            dec_packet = ZigbeeAppCommandPayload(bytes(packet.data))
            self.mark_packet(packet)
            self.transport_key = dec_packet.key
            self.complete()

    @property
    def output(self):
        return {
            "transport_key":self.transport_key
        }


class TouchlinkKeyManager(TrafficAnalyzer):
    def __init__(self, encrypted_key=None, unencrypted_key=None, transaction_id=None, response_id=None, key_index=None):
        self.reset()
        self.transaction_id = transaction_id
        self.response_id = response_id
        self.key_index = key_index
        self._encrypted_key = encrypted_key
        self._unencrypted_key = unencrypted_key

    def process_packet(self, packet):

        if ZLLScanRequest in packet:
            self.trigger()
            self.transaction_id = packet.inter_pan_transaction_id
            self.mark_packet(packet)
        elif ZLLScanResponse in packet:
            self.trigger()
            self.response_id = packet.response_id
            self.mark_packet(packet)
        elif ZLLNetworkJoinRouterRequest in packet:
            self.trigger()
            self.key_index = packet.key_index
            self._encrypted_key = packet.encrypted_network_key#.to_bytes(16, "big")
            self.mark_packet(packet)
            self.complete()

    @property
    def output(self):
        return {
            "key_index" : self.key_index,
            "encrypted_key": self._encrypted_key,
            "decrypted_key": self._decrypt_key(),
        }
    @property
    def encrypted_key(self):
        if self._encrypted_key is not None:
            return self._encrypted_key
        elif (
                self.transaction_id is not None and
                self.response_id is not None and
                self.key_index is not None and
                self._unencrypted_key is not None
        ):
            return self._encrypt_key()
        else:
            return None

    @property
    def unencrypted_key(self):
        if self._unencrypted_key is not None:
            return self._unencrypted_key
        elif (
                self.transaction_id is not None and
                self.response_id is not None and
                self.key_index is not None and
                self._encrypted_key is not None
        ):
            return self._decrypt_key()
        else:
            return None

    def reset(self):
        super().reset()
        self.transaction_id = None
        self.response_id = None
        self.key_index = None
        self._unencrypted_key = None
        self._encrypted_key = None


    def _encrypt_key(self):
        # If key index equals to zero, the key is encrypted using one single step with a key generated from transaction_id and response_id
        # (see Zigbee Cluster Specification v1.0 rev.6, section 13.3.4.10.4)
        if self.key_index == 0:
            # Generate the key
            development_key = b"PhLi" + pack(">I",self.transaction_id) + b"CLSN" + pack(">I",self.response_id)
            # Encrypt the unencrypted_key
            encryptor = AES.new(development_key, mode=AES.MODE_ECB)
            return encryptor.encrypt(self._unencrypted_key)
        else:
            # Select the right main key according to key index
            if self.key_index == 4:
                main_key = ZIGBEE_ZLL_KEYS["master_key"]
            elif self.key_index == 15:
                main_key = ZIGBEE_ZLL_KEYS["certification_key"]
            else:
                # No key found, return None and don't process further
                return None

            # Generate the plaintext vector
            plaintext = pack(">I",self.transaction_id)*2 + pack(">I",self.response_id)*2
            # Generate transport key
            step1 = AES.new(main_key, mode=AES.MODE_ECB)
            transport_key = step1.encrypt(plaintext)
            # Use transport key to decrypt network key
            step2 = AES.new(transport_key, mode=AES.MODE_ECB)
            return step2.encrypt(self._unencrypted_key)


    def _decrypt_key(self):
        # If key index equals to zero, the key is decrypted using one single step with a key generated from transaction_id and response_id
        # (see Zigbee Cluster Specification v1.0 rev.6, section 13.3.4.10.4)
        if self.key_index == 0:
            # Generate the key
            development_key = b"PhLi" + pack(">I",self.transaction_id) + b"CLSN" + pack(">I",self.response_id)
            # Decrypt the encrypted_key
            decryptor = AES.new(development_key, mode=AES.MODE_ECB)
            return decryptor.decrypt(self._encrypted_key)

        # If key not equals to zero, use a two-step decryption algorithm with known key
        # (see Zigbee Cluster Specification v1.0 rev.6, section 13.3.4.10.5)
        else:
            # Select the right main key according to key index
            if self.key_index == 4:
                main_key = ZIGBEE_ZLL_KEYS["master_key"]
            elif self.key_index == 15:
                main_key = ZIGBEE_ZLL_KEYS["certification_key"]
            else:
                # No key found, return None and don't process further
                return None

            # Generate the plaintext vector
            plaintext = pack(">I",self.transaction_id)*2 + pack(">I",self.response_id)*2
            # Generate transport key
            step1 = AES.new(main_key, mode=AES.MODE_ECB)
            transport_key = step1.encrypt(plaintext)
            # Use transport key to decrypt network key
            step2 = AES.new(transport_key, mode=AES.MODE_ECB)
            return step2.decrypt(self._encrypted_key)



"""
Bluetooth Low Energy Central connector
======================================

The :py:class:`whad.ble.connector.central.Central` class provides its own
asynchronous event notification mechanism (different from the one used in
the :py:class:`whad.ble.connector.peripheral.Peripheral` class for some
weird reasons).

"""
import logging
from time import sleep
from threading import Event
from typing import Optional

from whad.device.connector import Event as CentralEvent
from whad.common.stack import Layer
from whad.exceptions import UnsupportedCapability
from whad.hub.ble import Direction
from whad.hub.ble.bdaddr import BDAddress

from .base import BLE
from ..exceptions import ConnectionLostException, PeripheralNotFound
from ..profile.device import PeripheralDevice
from ..stack import BleStack
from ..stack.constants import BT_MANUFACTURERS, BT_VERSIONS
from ..stack.gatt import GattClient
from ..stack.att import ATTLayer
from ..stack.smp import CryptographicDatabase

logger = logging.getLogger(__name__)

class CentralConnected(CentralEvent):
    """Central connection event
    """

    def __init__(self, handle: int, local: BDAddress, remote: BDAddress):
        self.__handle = handle
        self.__local = local
        self.__remote = remote

    @property
    def handle(self) -> int:
        """Connection handle
        """
        return self.__handle

    @property
    def local_peer(self) -> BDAddress:
        """Central BD address
        """
        return self.__local

    @property
    def remote_peer(self) -> BDAddress:
        """Remote device BD address
        """
        return self.__remote

    def __repr__(self) -> str:
        return f"<CentralConnected handle={self.__handle} local={self.__local} remote={self.__remote}/>"

class CentralDisconnected(CentralEvent):
    """Client has disconnected from central
    """
    def __init__(self, handle: int):
        self.__handle = handle

    @property
    def handle(self) -> int:
        """Connection handle"""
        return self.__handle

class Central(BLE):
    """This connector provides a BLE Central role.

    To initiate a connection to a device, just call :meth:`Central.connect` with the target
    BD address and it should return an instance of 
    :class:`whad.ble.profile.device.PeripheralDevice` in return.
    """

    def __init__(self, device, existing_connection = None, from_json=None, stack=BleStack,
                 client=GattClient, security_database=None):
        """Attach a GATT client if specified in parameter

        If `client` is set to None, the default GATT layer is used, which does
        not provide any client feature at all.
        """
        super().__init__(device)

        # Reconfigure stack layers
        self.__configure_stack(stack, client)

        self.__gatt_client = None
        self.__connected = False
        self.__peripheral = None
        self.__profile_json = from_json
        self.__target = None
        self.__local = None
        self.__conn_handle = None
        self.connection = None

        # Internal events
        self.__connected_evt = Event()

        # If no connection, check if
        if not self.can_be_central():
            raise UnsupportedCapability('Central')

        # Initialize security database
        if security_database is None:
            logger.info((
                "No security database provided to this Peripheral instance, "
                "use a default one."))
            self.__security_database = CryptographicDatabase()
        else:
            logger.info('Peripheral will use the provided security database.')
            self.__security_database = security_database

        # Create our event handler
        self.add_event_handler(self.on_central_event)

        # If a connection already exists, just feed the stack with the parameters
        if existing_connection is not None:
            self.on_connected(existing_connection)
        else:
            # self.stop() # ButteRFly doesn't support calling stop when spawning central
            self.enable_central_mode()


    def __configure_stack(self, phy_layer=None, gatt_layer=None):
        """
        """
        # Save GATT and PHY layers
        if gatt_layer is not None:
            self.__gatt_layer = gatt_layer
        if phy_layer is not None:
            self.__phy_layer = phy_layer

            # Configure BLE stack to use our PHY class
            self.__stack = phy_layer(self)

        # Configure ATT layer to use our GATT class
        if self.__gatt_layer is not None:
            if issubclass(self.__gatt_layer, Layer) and self.__gatt_layer.alias == 'gatt':
                ATTLayer.add(self.__gatt_layer)

    @property
    def security_database(self):
        """Central role security database.
        """
        return self.__security_database

    @property
    def local_peer(self) -> BDAddress:
        """Local peer BD address.
        """
        return self.__local

    @property
    def target_peer(self) -> BDAddress:
        """Remote peer BD address.
        """
        return self.__target

    @property
    def conn_handle(self) -> int:
        """Connection handle.
        """
        return self.__conn_handle

    @property
    def stack(self):
        '''Return the current stack instance
        '''
        return self.__stack

    def set_profile_json(self, profile):
        """Set Central profile as json
        """
        self.__profile_json = profile

    def stop(self):
        """Stopping Central connector.
        """
        super().stop()

    def add_event_handler(self, handler, event_type = None):
        """Add an event handler for a specific event.
        """
        self.add_listener(handler, event_type)

    def remove_event_handler(self, handler):
        """Remove an event handler.
        """
        self.remove_listener(handler)

    def clear_event_handlers(self):
        """Remove all event handlers
        """
        self.clear_listeners()

    def on_central_event(self, event: CentralEvent):
        """Central event handler

        :param  event: Central event to handle
        :type   event: CentralEvent
        """

    def notify(self, event: CentralEvent):
        """Send event to handlers.

        This method is called from methods associated with incoming events
        and is called from the Central's connector IoThread

        :param  event: Central event to notify
        :type   event: CentralEvent
        """
        super().notify(event)

    def connect(self, bd_address, random=False, timeout=30, access_address=None,channel_map=None,
                crc_init=None, hop_interval=None, hop_increment=None) -> PeripheralDevice:
        """Connect to a target device

        :param  bd_address:     Bluetooth device address (in format 'xx:xx:xx:xx:xx:xx')
        :type   bd_address:     str
        :param  timeout:        Connection timeout
        :type   timeout:        float
        :param  access_address: Access address to use (optional)
        :type   access_address: int
        :param  channel_map:    Channel map to use (optional)
        :type   channel_map:    int
        :param  crc_init:       CRC Initialization value to use (optional)
        :type   crc_init:       int
        :param  hop_interval:   Hop interval to use (optional)
        :type   hop_interval:   int
        :param  hop_increment:  Hop increment to use (optional)
        :type   hop_increment:  int

        :return: An instance of `PeripheralDevice` on success, `None` on failure.
        :rtype: :class:`whad.ble.profile.device.PeripheralDevice`
        """
        # Clear connection event
        self.__connected_evt.clear()

        # Make sure our BLE stack is correctly configured
        self.__configure_stack()

        if self.can_connect():
            self.connect_to(
                bd_address,
                random=random,
                access_address=access_address,
                channel_map=channel_map,
                crc_init=crc_init,
                hop_interval=hop_interval,
                hop_increment=hop_increment
            )

            # Wait for connection to be established
            self.__connected_evt.wait(timeout=timeout)
            if self.__peripheral is None:
                raise PeripheralNotFound()

            # Return peripheral
            return self.peripheral()

        # Raise error
        raise PeripheralNotFound()

    def peripheral(self) -> Optional[PeripheralDevice]:
        """Connected BLE peripheral.
        """
        return self.__peripheral


    def send_pdu(self, pdu, conn_handle=0, direction=Direction.MASTER_TO_SLAVE,
                 access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the connected peripheral device or to the central device.

        :param  pdu:            BLE PDU to send.
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Direction (central to peripheral, peripheral to central)
        :type   direction:      :class:`whad.protocol.ble.ble_pb2.BleDirection`
        :param  access_address: Access address to use while sending PDU.
        :type   access_address: int
        :param  encrypt:        Enable PDU encryption if set to ``True``.
        :type   encrypt:        bool

        :return:                PDU transmission result.
        :rtype:                 bool
        """
        if self.__connected:
            return super().send_pdu(pdu, conn_handle, direction=direction,
                                    access_address=access_address, encrypt=encrypt)

        # Connection lost
        raise ConnectionLostException(None)

    ##############################
    # Incoming events
    ##############################

    def is_connected(self) -> bool:
        """Determine if the central device is connected to a peripheral.

        :return:    ``True`` if central is connected to a peripheral device, `False` otherwise.
        :rtype:     bool
        """
        return self.__connected

    def on_connected(self, connection_data):
        """Callback method to handle connection event.

        :param  connection_data: Connection data
        :type   connection_data: dict
        """
        # Save local and target peer info
        self.__local = BDAddress.from_bytes(
            connection_data.initiator,
            addr_type=connection_data.init_addr_type
        )

        self.__target = BDAddress.from_bytes(
            connection_data.advertiser,
            connection_data.adv_addr_type
        )
        self.__stack.on_connection(
            connection_data.conn_handle,
            self.__local,
            self.__target
        )
        self.__conn_handle = connection_data.conn_handle
        # Notify event handlers
        self.notify(CentralConnected(
            connection_data.conn_handle, self.__local,self.__target)
        )


    def on_disconnected(self, disconnection_data):
        """Callback method to handle disconnection event.

        :param  disconnection_data: Disconnection data
        :type   disconnection_data: :class:`whad.protocol.ble_pb2.Disconnected`
        """
        logger.debug("[central] Device disconnected, forward to stack")
        self.__stack.on_disconnection(
            disconnection_data.conn_handle,
            disconnection_data.reason
        )

        self.__connected = False
        self.__local = None
        self.__target = None

        # Notify peripheral device about this disconnection
        logger.debug("[central] Notify peripheral object that a disconnection occurred")
        if self.__peripheral is not None:
            self.__peripheral.on_disconnect(disconnection_data.conn_handle)

        # Notify event handlers
        self.notify(CentralDisconnected(self.conn_handle))


    def on_ctl_pdu(self, pdu):
        """This callback method is called whenever a control PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Central devices act as master, so we only forward slave to master
        messages to the stack.

        :param  pdu: BLE Control PDU
        :type   pdu: :class:`scapy.layers.bluetooth4LE.BTLE`
        """
        logger.info('received control PDU')
        if pdu.metadata.direction == Direction.SLAVE_TO_MASTER:
            self.__stack.on_ctl_pdu(pdu.metadata.connection_handle, pdu)

    def on_data_pdu(self, pdu):
        """This callback method is called whenever a data PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Central devices act as master, so we only forward slave to master
        messages to the stack.

        :param  pdu: BLE Data PDU
        :type   pdu: :class:`scapy.layers.bluetooth4LE.BTLE_DATA`
        """
        if pdu.metadata.direction == Direction.SLAVE_TO_MASTER:
            self.__stack.on_data_pdu(pdu.metadata.connection_handle, pdu)


    def on_new_connection(self, connection):
        """On new connection, discover primary services.

        :param  connection: New connection Protobuf message
        :type   connection: :class:`whad.protocol.ble_pb2.Connected`
        """
        logger.info('new connection established')

        # Use GATT client
        self.connection = connection
        self.__peripheral = PeripheralDevice(
            self,
            connection.gatt,
            connection.conn_handle,
            from_json=self.__profile_json
        )

        # Retrieve GATT client
        self.__gatt_client = connection.gatt
        self.__gatt_client.set_client_model(self.__peripheral)
        self.__connected = True
        self.__connected_evt.set()

        # Configure SMP layer
        # we set the security database
        self.connection.smp.set_security_database(self.__security_database)

        # Check if we got a matching LTK
        crypto_material = self.security_database.get(address=self.__target)

        if crypto_material is not None and crypto_material.has_ltk():
            self.__stack.get_layer('ll').state.register_encryption_key(
                connection.conn_handle,
                crypto_material.ltk.value
            )
            if crypto_material.is_authenticated():
                #print("Marked as authenticated")
                self.__stack.get_layer('ll').state.mark_as_authenticated(connection.conn_handle)
            else:
                pass#print("Marked as unauthenticated")

        # Notify peripheral about this connection
        self.__peripheral.on_connect(self.connection.conn_handle)


    def version(self, synchronous=True):
        """Query BLE version of remote peer.
        """
        if self.connection is not None:
            # Send an LL_VERSION_IND PDU
            self.connection.send_version()

            # Wait for an answer (mandatory)
            if synchronous:
                while not self.connection.remote_version:
                    sleep(0.01)
                result = self.connection.remote_version

                # Identify BT version
                if result.version in BT_VERSIONS:
                    version = BT_VERSIONS[result.version]
                else:
                    version = result.version

                # Identify BT company
                if result.company in BT_MANUFACTURERS:
                    company = BT_MANUFACTURERS[result.company]
                else:
                    company = result.company

                # Return information
                return (version, result.subversion, company)

        # Error
        return None

    def export_profile(self):
        """Export GATT profile of the existing connection.

        :return: Profile as a JSON string
        :rtype: str
        """
        return self.connection.gatt.model.export_json()

    def on_mtu_changed(self, conn_handle: int, mtu: int):
        """Notify MTU change to peripheral.
        """
        logger.info("on_mtu_changed")
        if self.__peripheral is not None:
            logger.debug("[Central::on_mtu_changed] MTU changed to %d, notify peripheral", mtu)
            self.__peripheral.on_mtu_changed(mtu)

    def set_mtu(self, mtu: int):
        """Set connection MTU.
        """
        if self.__gatt_client is not None:
            # Start a MTU exchange procedure
            self.__gatt_client.set_mtu(mtu)

    def get_mtu(self) -> Optional[int]:
        """Retrieve the connection MTU.
        """
        if self.connection is not None:
            return self.connection.att.get_server_mtu()



from dataclasses import dataclass, field

@dataclass
class SnifferConfiguration:
    """
    Configuration for sniffing a 802.15.4 communication.

    :param channel: select the channel to sniff (c)

    """
    channel : int = 11



"""
WHAD install tool

This utility is a simple helper facilitating the install of devices.
"""
import logging
import os
import sys
import stat
import shlex
import inspect
from typing import Union
from subprocess import Popen, PIPE
from shutil import copy, which
from pathlib import Path
from grp import getgrall

import usb
import requests
import distro

from serial.tools.list_ports import comports
from prompt_toolkit import print_formatted_text, HTML

import whad
from whad.cli.ui import info, error, warning, success
from whad.cli.app import CommandLineApp, ApplicationError



logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadInstallApp(CommandLineApp):
    """winstall main CLI application class.
    """

    def __init__(self):
        """Application uses no interface and has no commands.
        """
        super().__init__(
            description='WHAD install tool',
            interface=False,
            commands=False
        )


        self.add_argument(
            '--list',
            '-l',
            dest='list',
            action="store_true",
            default=False,
            help='List existing devices'
        )

        self.add_argument(
            '--rules',
            '-r',
            dest='rules',
            action="store_true",
            default=False,
            help='Install rules for selected devices'
        )


        self.add_argument(
            '--port',
            '-p',
            dest='port',
            default=False,
            help='Provide a specific port to use during flash'
        )

        self.add_argument(
            '--flash',
            '-f',
            dest='flash',
            action="store_true",
            default=False,
            help='Flash latest firmwares for selected devices'
        )

        self.add_argument(
            'device',
            nargs="*",
            default=[],
            help='Select a specific device'
        )

        self.need_reload = False
        self.udev_rules = None

    def check_tools_availability(self, *tools):
        """Check that a list of tools are installed on the system.
        """
        return all([which(tool) is not None for tool in tools])

    def ask_for_privileges(self):
        """Ask user to get root privileges through sudo.
        """
        euid = os.geteuid()
        if euid != 0:
            info("This tool must run as root, let's elevate your privileges !")
            args = ['sudo', sys.executable] + sys.argv + [os.environ]
            # the next line replaces the currently-running process with the sudo
            pypath = ":".join(sys.path)
            args = ['sudo', f"PYTHONPATH={pypath}", sys.executable] + sys.argv
            os.execvpe('sudo', args, os.environ)
        return os.geteuid() == 0

    def get_udev_rule_filename(self, device_name: str) -> str:
        """Return udev rule filepath for given device.

        :param device_name: Device name
        :type device_name: str
        :return: udev rule filepath
        :rtype: str
        """
        return os.path.realpath(
            f"{os.path.dirname(whad.__file__)}/resources/rules/{device_name}.rules"
        )


    def install_udev_rule(self, device_name: str) -> bool:
        """Deploy udev rule for specified device.

        :param device_name: Device name
        :type device_name: str
        :return: `True` if udev rule install has been successful, `False` otherwise.
        :rtype: bool
        """
        matching_rules = [udev_name for udev_name in self.udev_rules if device_name in udev_name]
        if len(matching_rules) == 0:
            rule_filename = self.get_udev_rule_filename(device_name)
            dev_rule = "/usr/lib/udev/rules.d/40-"+device_name+".rules"
            if os.path.exists(rule_filename):
                copy(rule_filename, dev_rule)
                success(f"Rule successfully installed for device '{device_name}': {dev_rule}")
                return True

            # Failed to install rule.
            warning("Rule not found, skipping.")
            return False

        # Already installed.
        dev_rules = ",".join(matching_rules)
        warning(f"Rules detected for device '{device_name}' ({dev_rules}), skipping.")
        return False

    def get_latest_release(self, github_repository: str) -> Union[str, None]:
        """Retrieve the latest release for a given github repository.

        :param github_repository: Target github repository URL
        :type github_repository: str
        :return: URL to latest release if found, `None` otherwise.
        :rtype: str
        """
        g_request = requests.get(f"https://github.com/{github_repository}/releases/latest")
        if g_request.status_code == 200:
            return g_request.url

        # Failed.
        return None

    def install_nrfutil(self):
        """Install nRFutil for Linux.
        """
        info("Downloading 'nrfutil' tool from nordicsemi.com...")
        nrfutil_url = (
            "https://files.nordicsemi.com/ui/api/v1/download?repoKey=swtools"
            "&path=external/nrfutil/executables/x86_64-unknown-linux-gnu/nrfutil"
        )
        n_request = requests.get(nrfutil_url, allow_redirects=True)
        if n_request.status_code == 200:
            # Remove if already existing
            if os.path.exists("/tmp/nrfutil"):
                os.unlink("/tmp/nrfutil")

            # Save nrfutil to temporary file
            with open("/tmp/nrfutil", "wb") as nrfutil:
                nrfutil.write(n_request.content)

            # Change permissions to allow execution
            f = Path("/tmp/nrfutil")
            f.chmod(f.stat().st_mode | stat.S_IEXEC)

            # Use nrfutil to install nRF SDK and tools
            _, _, e = self.run_command('/tmp/nrfutil install nrf5sdk-tools --force')
            if len(e) > 0:
                return False
            return True

        # Cannot download nrfutil
        return False

    def download_latest_butterfly_release(self):
        """Download latest ButteRFly release from github.
        """
        latest_release_url = self.get_latest_release("whad-team/butterfly")
        try:
            release = latest_release_url.split("/tag/")[1]
            success(f"Latest butterfly release: {release}")
            zip_url = (
                f"https://github.com/whad-team/butterfly/releases/download/"
                f"{release}"
                f"/butterfly-fwupgrade.zip"
            )

            # Download latest release from github
            z_request = requests.get(zip_url, allow_redirects=True)
            if z_request.status_code == 200:
                latest_release_filename = f"/tmp/butterfly-{release}.zip"

                # Remove if previously downloaded
                if os.path.exists(latest_release_filename):
                    os.unlink(latest_release_filename)

                # Save to temporary file
                with open(latest_release_filename, "wb") as fw:
                    fw.write(z_request.content)

                # Return file path
                return latest_release_filename

        except IndexError:
            error("Release not found, exiting...")

        # Fail downloading and installing ButteRFly.
        return None

    def flash_firmware_with_nrfutil(self, firmware, serial_port="/dev/ttyACM0",
                                    nrfutil_exec="/tmp/nrfutil") -> bool:
        """Flash a compatible nRF device with a specified firmware thanks to nrfutil.

        :param firmware: Firmware to flash
        :type firmware: str
        :param serial_port: Target device port
        :type serial_port: str
        :param nrfutil_exec: Path to nrfutil binary
        :type nrfutil_exec: str
        :return: `True` if device has been successfully flashed with specified firmware,
                 `False` otherwise.
        :rtype: bool
        """
        _, o, _  = self.run_command(
            f"{nrfutil_exec} dfu usb-serial -pkg {firmware}  -p {serial_port} -b 115200"
        )

        return b"Device programmed" in o

    def flash_butterfly(self, port="/dev/ttyACM0"):
        """Flash ButteRFly on a compatible device.
        """
        if not self.install_nrfutil():
            error("Failure during 'nrfutil' installation.")
            return False

        latest_release_filename = self.download_latest_butterfly_release()
        if latest_release_filename is None:
            error("Failure during 'butterfly' release download")
            return False

        result = self.flash_firmware_with_nrfutil(
            latest_release_filename, serial_port=port, nrfutil_exec="/tmp/nrfutil"
        )
        return result

    def run_command(self, cmd: str, shell: bool = False, force_env: dict = None,
                    cwd: str = None, added_path: str = None):
        """Run the specified command and return its output.

        :param cmd: Command to execute
        :type cmd: str
        :param shell: Execute command in shell if set to `True`
        :type shell: bool, optional
        :param force_env: Specify a custom environment to use
        :type force_env: dict, optional
        :param cwd: Specify the current working directory
        :type cwd: str, optional
        :param added_path: Add specific string to PATH
        :type added_path: str, optional
        :rtype: tuple
        :return: A tuple containing the return code and the textual output
        """
        info("Running command: " + cmd)

        if added_path is not None:
            env  = os.environ.copy()
            env["PATH"] = f"{env['PATH']}:/" + added_path
        elif force_env is not None:
            env = force_env
        else:
            env = os.environ

        with Popen(shlex.split(cmd) if not shell else cmd,  stderr=PIPE,
                        stdout=PIPE, shell=shell, cwd=cwd, env=env) as process:
            stdout, stderr = process.communicate()

            if process.returncode != 0:
                error("Standard output:" + stdout.decode())
                error("Standard error:" + stderr.decode())

            return process.returncode , stdout, stderr


    def install_ubertooth_host(self):
        """Install Ubertooth tools and libraries.
        """
        if self.check_tools_availability('ubertooth-dfu'):
            return True

        distribution = distro.id()

        if distribution == 'fedora':
            r, _, _ = self.run_command('dnf install -y ubertooth')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt install -y ubertooth')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        return True


    def download_latest_ubertooth_release(self) -> Union[str, None]:
        """Download latest version of Ubertooth firmware.
        
        :return: Firmware file path
        :rtype: str
        """
        latest_release_url = self.get_latest_release("greatscottgadgets/ubertooth")
        try:
            release = latest_release_url.split("/tag/")[1]
            success(f"Latest ubertooth release: {release}")
            zip_url = (
                f"https://github.com/greatscottgadgets/ubertooth/releases/download/"
                f"{release}/ubertooth-{release}.tar.xz"
            )
            z_request = requests.get(zip_url, allow_redirects=True)
            if z_request.status_code == 200:
                latest_release_filename = "/tmp/ubertooth-{release}.tar.xz"

                # Remove previously downloaded firmware if any
                if os.path.exists(latest_release_filename):
                    os.unlink(latest_release_filename)

                # Save latest release in file
                with open(latest_release_filename, "wb") as fw:
                    fw.write(z_request.content)

                # Return firmware file path
                return latest_release_filename
        except IndexError:
            error("Release not found, exiting...")

        # Failed to download latest release
        return None

    def flash_ubertooth(self, port=None):
        """Flash an Ubertooth device with the latest Ubertooth firmware.
        """
        if not self.install_ubertooth_host():
            error("Failure during 'ubertooth-dfu' installation.")
            return False

        latest_release_filename = self.download_latest_ubertooth_release()
        r, _, _ = self.run_command(f"tar -xf {latest_release_filename}", cwd="/tmp")
        if r != 0:
            return False

        port_option = f"-U {port}" if port is not None else ""
        r, _, _ = self.run_command(
            f"ubertooth-dfu -d bluetooth_rxtx.dfu -r {port_option}",
            cwd=latest_release_filename.replace(".tar.xz","/ubertooth-one-firmware-bin/")
        )
        if r != 0:
            return False
        return True


    def flash_yardstickone(self, port="/dev/RFCAT_BL_YS1"):
        """Flash a YardStickOne with the latest firmware release.
        """
        r, _, _ = self.run_command("rm -rf /tmp/rfcat")
        if r != 0:
            return False

        r, _, _ = self.run_command("git clone https://github.com/atlas0fd00m/rfcat", cwd="/tmp")
        if r != 0:
            return False

        r, _, _ = self.run_command("pip install future")
        if r != 0:
            return False


        warning((
            "For some reason, latest build from rfcat repo crashes the dongle, "
            "let's get the latest functional build."
        ))

        r, _, _ = self.run_command((
            "wget https://gist.githubusercontent.com/mossmann/7b816680df2ac513df3835f3cb9eaa1b"
            "/raw/2f33a76a86c8f6ee71dacb385537386486fe633a/RfCatYS1CCBootloader.hex -O /tmp/ys1.hex"
        ))
        if r != 0:
            return False

        if not os.path.exists(port):
            r, _, _ = self.run_command("/tmp/rfcat/rfcat --bootloader --force")
            if r != 0:
                return False
        r, _, _ = self.run_command(
            f"/tmp/rfcat/CC-Bootloader/rfcat_bootloader {port} erase_all"
        )
        if r != 0:
            return False
        r, _, _ = self.run_command(
            f"/tmp/rfcat/CC-Bootloader/rfcat_bootloader {port} download /tmp/ys1.hex"
        )
        if r != 0:
            return False
        r, _, _  = self.run_command(
            f"/tmp/rfcat/CC-Bootloader/rfcat_bootloader {port} verify /tmp/ys1.hex"
        )
        if r != 0:
            return False
        r, _, _ = self.run_command(
            f"/tmp/rfcat/CC-Bootloader/rfcat_bootloader {port} run"
        )
        if r != 0:
            return False
        return True

    def install_sdcc(self):
        """Install SDCC.
        """
        if self.check_tools_availability("sdcc"):
            return True

        distribution = distro.id()

        if distribution == "fedora":

            r, _, _ = self.run_command("rm -rf /usr/bin/packihx /usr/bin/sdcc")
            if r != 0:
                return False

            r, _, _ = self.run_command("dnf install -y sdcc")
            if r != 0:
                return False

            r, _, _ = self.run_command("ln -s /usr/bin/sdcc-sdcc /usr/bin/sdcc")
            if r != 0:
                return False

            r, _, _ = self.run_command("ln -s /usr/bin/sdcc-packihx /usr/bin/packihx")
            if r != 0:
                return False

        elif distribution in ("debian", "ubuntu", "raspbian"):
            r, _, _ = self.run_command("apt install -y sdcc")
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        return True

    def flash_rfstorm(self):
        """Flash a compatible device with Bastille's RFStorm firmware.
        """
        if not self.install_sdcc():
            error("An error occurred during 'sdcc' installation.")

        r, _, _ = self.run_command("rm -rf /tmp/nrf-research-firmware-python3")
        if r != 0:
            return False
        r, _, _ = self.run_command("pip install pyusb platformio")
        if r != 0:
            return False
        r, _, _ = self.run_command(
            "git clone https://github.com/kuzmin-no/nrf-research-firmware-python3",
            cwd="/tmp"
        )
        if r != 0:
            return False
        r, _, _ = self.run_command("make", cwd="/tmp/nrf-research-firmware-python3")
        if r != 0:
            return False
        r, o, _ = self.run_command("lsusb")
        if b"1915:0102" in o or b"046d:c52b" in o:
            r, _, _ = self.run_command(
                "make logitech_install",
                cwd="/tmp/nrf-research-firmware-python3"
            )
            return r == 0

        r, _, _ = self.run_command(
            "make install", cwd="/tmp/nrf-research-firmware-python3"
        )
        return r == 0

    def install_esp_sdk(self):
        """Install Espressif SDK (esp-idf).
        """
        if self.check_tools_availability("idf.py"):
            return True

        distribution = distro.id()
        if distribution == "fedora":
            r, _, _ = self.run_command((
                "dnf install -y wget flex bison gperf python3 cmake ninja-build"
                " ccache dfu-util libusbx"
            ))
            if r != 0:
                return False

        elif distribution in ("debian", "ubuntu", "raspbian"):
            r, _, _ = self.run_command((
                "apt-get install -y git wget flex bison gperf python3 python3-pip"
                " python3-venv cmake ninja-build ccache libffi-dev libssl-dev dfu-util "
                "libusb-1.0-0"
            ))
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        r, _, _ = self.run_command("rm -rf /tmp/esp")
        if r != 0:
            return False
        r, _, _ = self.run_command("mkdir -p /tmp/esp")
        if r != 0:
            return False

        r, _, _ = self.run_command(
            "git clone -b release/v4.4 --recursive https://github.com/espressif/esp-idf.git",
            cwd="/tmp/esp"
        )
        if r != 0:
            return False

        # Clear python path
        env = os.environ.copy()
        env["PYTHONPATH"] = ""

        r, _, _ = self.run_command("./install.sh esp32", shell=True, force_env=env,
                                   cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        r, o, _ = self.run_command(". ./export.sh > /dev/null; env", shell=True,
                                   force_env=env, cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        os.environ.update(dict(
            line.decode().partition('=   ')[::2] for line in o.split(b'\n')[:-1]
        ))

        r, o, _ = self.run_command("idf.py --version", force_env=os.environ,
                                   cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        success("Espressif framework installed: " + o.decode())
        return True


    def flash_esp(self, port=None):
        """Flash an ESP device with latest NodeMCU WHAD firmware.
        """
        if not self.install_esp_sdk():
            error("An error occurred uring ESP SDK installation.")
            return False

        info("Cloning 'nodemcu-esp32-firmware' github repository...")
        r, _, _ = self.run_command("rm -rf /tmp/nodemcu-esp32-firmware/", force_env=os.environ,
                                   cwd="/tmp")
        if r != 0:
            return False

        r, _, _ = self.run_command(
            "git clone https://github.com/whad-team/nodemcu-esp32-firmware/ --recurse",
            force_env=os.environ,  cwd="/tmp"
        )
        if r != 0:
            return False

        info("Building firmware...")
        r, _, _ = self.run_command("idf.py build", force_env=os.environ,
                                   cwd="/tmp/nodemcu-esp32-firmware")
        if r != 0:
            return False

        info("Flashing firmware...")
        dev_port = f"-p {port}" if port is not None else ""
        r, _, _ = self.run_command(f"idf.py flash {dev_port}", force_env=os.environ,
                                   cwd="/tmp/nodemcu-esp32-firmware")
        if r != 0:
            return False

        # Success
        return True


    def download_latest_stm32wl55_release(self):
        """Download latest stable version of WHAD stm32wl55 compatible firmware
        for ST Nucleo WL55 board.
        """
        latest_release_url = self.get_latest_release("whad-team/stm32wlxx-firmware")
        try:
            release = latest_release_url.split("/tag/")[1]
            success("Latest STM32WL55 firmware release: " + release)
            zip_url = (
                f"https://github.com/whad-team/stm32wlxx-firmware/releases/download/"
                f"{release}/nucleo_wl55.hex"
            )
            z_request = requests.get(zip_url, allow_redirects=True)
            if z_request.status_code == 200:
                latest_release_filename = "/tmp/nucleo_wl55.hex"

                # Remove previous file if any
                if os.path.exists(latest_release_filename):
                    os.unlink(latest_release_filename)

                # Store latest version in temporary directory
                with open(latest_release_filename, "wb") as fw:
                    fw.write(z_request.content)

                # Return firmware path
                return latest_release_filename
        except IndexError:
            error("Release not found, exiting...")

        # Download failed
        return None

    def download_latest_lorae5mini_release(self):
        """Download the latest version of WHAD compatible firmware for SeeedStudio's
        LoRa-e5-mini/Wio-e5-mini board.
        """
        latest_release_url = self.get_latest_release("whad-team/stm32wlxx-firmware")
        try:
            release = latest_release_url.split("/tag/")[1]
            success("Latest STM32WL55 firmware release: " + release)
            zip_url = (
                f"https://github.com/whad-team/stm32wlxx-firmware/releases/download/"
                f"{release}/lora_e5_mini.hex"
            )
            z_request = requests.get(zip_url, allow_redirects=True)
            if z_request.status_code == 200:
                latest_release_filename = "/tmp/lora_e5_mini.hex"

                # Remove firmware if file exists
                if os.path.exists(latest_release_filename):
                    os.unlink(latest_release_filename)

                # Save latest version in temporary file
                with open(latest_release_filename, "wb") as fw:
                    fw.write(z_request.content)

                # Return firmware file path
                return latest_release_filename
        except IndexError:
            error("Release not found, exiting...")

        # Download failed.
        return None

    def install_stflash(self):
        """Install stflash.
        """
        if self.check_tools_availability('st-flash'):
            return True

        distribution = distro.id()
        if distribution == 'fedora':
            r, _, _ = self.run_command('dnf install -y stlink')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt-get install -y stlink-tools')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")

        # Cannot install st-flash.
        return False



    def flash_nucleowl55(self):
        """Flash a ST Nucleo WL55 devboard with compatible firmware.
        """
        if not self.install_stflash():
            error("An error occurred during 'st-flash' installation.")
            return False


        latest_release_filename = self.download_latest_stm32wl55_release()
        r, _, _ = self.run_command(f"st-flash --format=ihex write {latest_release_filename}",
                                   cwd="/tmp")
        if r != 0:
            return False

        return True

    def flash_lorae5mini(self):
        """Flash a SeeedStudio LoRa-e5-mini board with compatible firmware.
        """
        if not self.install_stflash():
            error("An error occurred during 'st-flash' installation.")
            return False


        latest_release_filename = self.download_latest_lorae5mini_release()
        r, _, _ = self.run_command(f"st-flash --format=ihex write {latest_release_filename}",
                                   cwd="/tmp")
        if r != 0:
            return False

        return True

    def rules_rfstorm(self):
        """Install udev rules for RFStorm firmware.
        """
        if self.install_udev_rule("rfstorm"):
            self.need_reload = True
        return True

    def rules_ubertooth(self):
        """Install udev rules for Ubertooth.
        """
        if self.install_udev_rule("ubertooth"):
            self.need_reload = True
        return True

    def rules_rzusbstick(self):
        """Install udev rules for AVR RZUSB Stick
        """
        if self.install_udev_rule("rzusbstick"):
            self.need_reload = True
        return True

    def rules_yardstickone(self):
        """Install udev rules for YardStickOne
        """
        if self.install_udev_rule("yardstickone"):
            self.need_reload = True
        return True

    def rules_nucleowl55(self):
        """Install udev rules for ST Nucleo WL55 devboard
        """
        return self.install_serial_port_capabilities()

    def rules_lorae5mini(self):
        """Install udev rules for SeeedStudio's LoRa-e5-mini
        """
        return self.install_serial_port_capabilities()

    def rules_esp(self):
        """Install udev rules for Espressif ESP devices
        """
        return self.install_serial_port_capabilities()

    def rules_butterfly(self):
        """Install udev rules for ButteRFly compatible hardware
        """
        return self.install_serial_port_capabilities()

    def rules_apimote(self):
        """Install udev rules for Apimote
        """
        return self.install_serial_port_capabilities()

    def rules_hci(self):
        """Install sniffing capabilities for HCI devices on Linux
        """
        python_interpreter = os.path.realpath(sys.executable)
        if not self.install_hci_capabilities(python_interpreter):
            return False
        success((f"HCI capabilities successfully added to python interpreter "
                f"({python_interpreter})."))
        return True

    def pre_run(self):
        """Pre-processing parameters.
        """
        super().pre_run()
        # find udev location
        candidate_udev_locations = [
            "/etc/udev/rules.d",
            "/lib/udev/rules.d",
            "/run/udev/rules.d",
            "/var/run/udev/rules.d"
        ]
        self.udev_rules = []

        for location in candidate_udev_locations:
            if os.path.exists(location):
                self.udev_rules += os.listdir(location)

        self.need_reload = False


    def install_serial_port_capabilities(self):
        """Configure machine to allow serial port access in user-mode.
        """
        if not self.check_tools_availability("usermod"):
            error("'usermod' tool is required but not available")
            return False
        # Find the right group which depends on the distribution (e.g. dialout
        # for Debian, uucp for Arch-Linux).
        groups = [ group.gr_name for group in getgrall() ]
        if "dialout" in groups:
            group = "dialout"
        elif "uucp" in groups:
            group = "uucp"
        else:
            error("Group granting serial port capabilities not found!")
            return False
        with Popen(['usermod', '-a', '-G', group, os.getlogin()], stdout=PIPE,
                   stderr=PIPE) as process:
            _, stderr = process.communicate()

            if len(stderr) > 0:
                return False

            success((
                f"Serial port capabilities successfully configured "
                f"(user {os.getlogin()} added to group {group}): logout to take "
                f"modification into account."
            ))

            # Success.
            return True


    def reload_udev_rules(self):
        """Reload udev rules on Linux host
        """
        if not self.check_tools_availability("udevadm"):
            error("'udevadm' tool is required but not available")
            return False
        info("Reloading udev rules...")
        with Popen(['udevadm', 'control', '--reload-rules'], stderr=PIPE,
                   stdout=PIPE) as process:
            _, stderr = process.communicate()
            if len(stderr) > 0:
                return False

        with Popen(['udevadm', 'trigger'],  stderr=PIPE, stdout=PIPE) as process:
            _, stderr = process.communicate()
            if len(stderr) > 0:
                return False

        return True


    def install_hci_capabilities(self, interpreter):
        """Install HCI user-mode access
        """
        if not self.check_tools_availability("getcap", "setcap"):
            error("'getcap' and 'setcap' tools are required but not available")
            return False

        with Popen(['getcap', interpreter], stdout=PIPE, stderr=PIPE) as process:
            stdout, stderr = process.communicate()

            if "cap_net_admin,cap_net_raw=eip" not in stdout.decode('utf-8'):
                with Popen(['setcap', 'cap_net_admin,cap_net_raw+eip',interpreter],
                                stdout=PIPE, stderr=PIPE) as setcap:
                    stdout, stderr = setcap.communicate()

                    if len(stdout) != 0 or len(stderr) != 0:
                        return False

            return True

        return True

    def list_devices(self):
        """Enumerate compatible devices
        """
        info("Detected devices:")
        print()
        for dev in comports():
            if (dev.vid == 0x10c4 and dev.pid == 0xea60) or dev.vid == 0x303a:
                print_formatted_text(HTML(f"  - <b>Espressif ESP-32 board: </b> {dev.device}"))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules esp</i>"
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash esp "
                    f"--port {dev.device}</i>"
                )))
                print()
            elif dev.vid == 0xc0ff and dev.pid == 0xeeee:
                print_formatted_text(HTML("  - <b>WHAD ButteRFly dongle: </b> {dev.device}"))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules butterfly</i>"
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash butterfly "
                    f"--port {dev.device}</i>"
                )))
                print()
            elif dev.vid == 0x0483 and dev.pid == 0x374e:
                print_formatted_text(HTML(f"  - <b>Nucleo STM32WL55: </b> {dev.device}"))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules nucleowl55</i>"
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash nucleowl55 "
                    f"--port {dev.device}</i>"
                )))
                print()

        ubertooth_count = 0
        yardstickone_count = 0
        rfstorm_count = 0
        rzusbstick_count = 0
        for device in usb.core.find(find_all=1):
            vid, pid = device.idVendor, device.idProduct
            if (vid == 0xffff and pid == 0x0004) or (vid == 0x1d50 and pid >= 0x6000 and pid <= 0x6003):
                print_formatted_text(HTML(
                    f"  - <b>Ubertooth One: </b> {ubertooth_count}"
                ))
                print_formatted_text(HTML(
                    (
                        "    <u>Command (install rules):</u> <i>winstall "
                        "--rules ubertooth </i>"
                    )
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash ubertooth "
                    f"--port {ubertooth_count}</i>"
                )))
                ubertooth_count += 1
                print()
            elif (vid == 0x1d50 and pid in (0x605b, 0x6047, 0x6048, 0xecc1, 0x6049,
                                            0x604a, 0x605c, 0xecc0)):
                print_formatted_text(HTML(
                    f"  - <b>Yard Stick One: </b> {yardstickone_count}"
                ))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules yardstickone </i>"
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash yardstickone "
                    f"--port {yardstickone_count}</i>"
                )))
                yardstickone_count += 1
                print()
            elif (vid == 0x046d and pid in (0xc52b, 0xaaaa)) or (vid == 0x1915 and pid in (0x0102, 0x7777, 0x0101)):
                print_formatted_text(HTML(
                    f"  - <b>RFStorm compatible dongle: </b> {rfstorm_count}"
                ))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules rfstorm </i>"
                ))
                print_formatted_text(HTML((
                    f"    <u>Command (flash firmware):</u> <i>winstall --flash rfstorm "
                    f"--port {rfstorm_count}</i>"
                )))
                rfstorm_count += 1
                print()
            elif (vid == 0x03eb and pid == 0x210a):
                print_formatted_text(HTML(
                    f"  - <b>RZUSBStick: </b> {rzusbstick_count}"
                ))
                print_formatted_text(HTML(
                    "    <u>Command (install rules):</u> <i>winstall --rules rzusbstick </i>"
                ))
                rzusbstick_count += 1
                print()

        for device in os.listdir("/sys/class/bluetooth"):
            print_formatted_text(HTML(f"  - <b>HCI device: </b> {device}"))
            print_formatted_text(HTML(
                "    <u>Command (install rules):</u> <i>winstall --rules hci </i>"
            ))
            print()

    def run(self):
        """winstall main function.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            rules = self.args.rules
            flash = self.args.flash

            if not self.args.rules and not self.args.flash:
                # By default, if no option is provided, only use rules
                self.args.list = True

            if self.args.list or len(self.args.device) == 0:
                self.list_devices()
                return

            # We need root, let's ask
            if not self.ask_for_privileges():
                error("Login failed, aborting.")
                sys.exit(1)

            if  "all" in self.args.device:
                self.args.device = [
                    "hci",
                    "butterfly",
                    "esp",
                    "lorae5mini",
                    "rfstorm",
                    "ubertooth",
                    "apimote",
                    "nucleowl55",
                    "rzusbstick",
                    "yardstickone"
                ]

            # Iterate over devices
            for device in self.args.device:
                if flash:
                    if hasattr(self, "flash_" + device):
                        info(f"Flashing {device} device ...")

                        flashfunc = getattr(self, "flash_" + device)
                        if self.args.port and "port" in list(inspect.signature(flashfunc).parameters):
                            if flashfunc(port=self.args.port):
                                success(f"Flashing successful for device '{device}'.")
                            else:
                                error((f"An error occurred during device flashing"
                                      f" for device '{device}'."))
                                sys.exit(1)
                        else:
                            if flashfunc():
                                success(f"Flashing successful for device '{device}'.")
                            else:
                                error((f"An error occurred during device flashing"
                                      f" for device '{device}'."))
                                sys.exit(1)
                    else:
                        warning(f"Flashing {device} device is not supported yet.")

                if rules:
                    if hasattr(self, f"rules_{device}"):
                        info(f"Installing rules for {device} device ...")
                        if getattr(self, f"rules_{device}")():
                            success(f"Rules successfully added for device '{device}'.")
                        else:
                            error((f"An error occurred during device rules installation "
                                  f"for device '{device}'."))
                            sys.exit(1)
                    else:
                        warning(f"Installing rules for {device} device is not supported yet.")


            if rules and self.need_reload:
                if not self.reload_udev_rules():
                    error("An error occurred during udev rules reload, exiting.")
                    sys.exit(1)
                else:
                    success("Rules successfully reloaded.")
        except Exception as e:
            error("An error occurred: " + repr(e))

        # Launch post-run tasks
        self.post_run()

def winstall_main():
    """Launcher for winstall
    """
    try:
        app = WhadInstallApp()
        app.run()
    except ApplicationError as err:
        err.show()



import json
import re


class InvalidBDAddressException(Exception):
    """Invalid BD address used
    """


class BDAddress(object):
    """This class represents a Bluetooth Device address.
    """

    PUBLIC = 0x00
    RANDOM = 0x01

    def __init__(self, address, random: bool = False, addr_type=None):
        """Initialize BD address

        By default, BD address is public unless `random` is set to True.

        :param int addr_type: Set BD address type (either BDAddress.PUBLIC or BDAddress.RANDOM). This setting overseeds `random`
        :param bool random: Set BD address as random if set to True. BD address is public by default.
        """
        if addr_type is not None:
            if addr_type in [BDAddress.PUBLIC, BDAddress.RANDOM]:
                self.__type = addr_type
        elif random:
            self.__type = BDAddress.RANDOM
        else:
            self.__type = BDAddress.PUBLIC

        if isinstance(address, str):
            if re.match(r"^([0-9a-fA-F]{2}\:){5}[0-9a-fA-F]{2}$", address) is not None:
                self.__value = bytes.fromhex(address.replace(":", ""))[::-1]
            elif re.match("[0-9a-fA-F]{12}$", address) is not None:
                self.__value = bytes.fromhex(address)[::-1]
            else:
                raise InvalidBDAddressException
        elif isinstance(address, bytes) and len(address) == 6:
            self.__value = address
        else:
            raise InvalidBDAddressException

    def __eq__(self, other):
        return (self.__value == other.value) and (self.__type == other.type)

    def __str__(self):
        return ":".join(f"{b:02x}" for b in self.__value[::-1])

    def __repr__(self):
        return f"BDAddress({str(self)})"

    def export_json(self) -> str:
        return json.dumps(str(self))

    @property
    def type(self) -> int:
        return self.__type

    @property
    def value(self) -> bytes:
        return self.__value

    def is_public(self) -> bool:
        """Determine if address is public.

        :rtype: bool
        :return: True if BD address is public, False otherwise.
        """
        return self.__type == BDAddress.PUBLIC

    def is_random(self) -> bool:
        """Determine if address is random.

        :rtype: bool
        :return: True if BD address is random, False otherwise.
        """
        return self.__type == BDAddress.RANDOM

    @staticmethod
    def from_bytes(bd_addr_bytes: bytes, addr_type=None) -> "BDAddress":
        """Convert a 6-byte array into a Bluetooth address.

        :param bytes bd_addr_bytes: Bluetooth Device address as bytes.
        :param addr_type: Bluetooth Device Address type, PUBLIC or RANDOM
        :type addr_type: int, optional
        :rtype: BDAddress
        :return: An instance of BDAddress representing the corresponding BD address.
        """
        if len(bd_addr_bytes) == 6:
            return BDAddress(bd_addr_bytes, addr_type=addr_type)
        else:
            raise InvalidBDAddressException

    @staticmethod
    def check(bd_addr: str) -> bool:
        """Check if a BD address is valid

        :param      bd_addr: BD address to test
        :type       bd_addr: str
        :return:    ``True`` if BD address is valid, ``False`` otherwise
        :rtype:     bool
        """
        return bool(re.match("^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$", bd_addr))



"""WHAD Protocol hub Registry

This module implements the Registry class used in automated protocol buffers
messages parsing.
"""
from typing import Optional
from scapy.packet import Packet

from .exceptions import UnsupportedVersionException

class Registry:
    """Protocol message versions registry.
    """

    VERSIONS = {}
    NAME = None

    @classmethod
    def add_node_version(parent_class, version: int, name: str, clazz):
        """Add a specific class `clazz` to our message registry for version
        `version` with alias `name`.
        """
        # If version is unknown, create it
        if version not in parent_class.VERSIONS:
            parent_class.VERSIONS[version] = {}

        # Add clazz based on provided alias for this version
        parent_class.VERSIONS[version][name] = clazz

    @classmethod
    def bound(cls, name: str = None, version: int = 1):
        """Retrieve the given node class `name` for version `version`.

        If there is no defined class for version N, look for a corresponding
        class in version N-1, N-2 until 0.
        """
        # Look for node class from given name and version
        if version in cls.VERSIONS:
            if name in cls.VERSIONS[version]:
                return cls.VERSIONS[version][name]

        if version > 1:
            # If not found for version N, look for node class in version N-1
            return cls.bound(name, version - 1)

        # If not found, raise exception
        raise UnsupportedVersionException(name, version)

    def is_packet_compat(self, packet: Packet) -> bool:
        """Default method to determine if a packet is compatible or not."""
        return False


    def convert_packet(self, packet) -> Optional['HubMessage']:
        """Default packet conversion method."""
        return None




"""LoRaWAN Channel Plan management

LoRaWAN Channel Plan
--------------------

LoRaWAN networks deployed in various regions of the globe don't use the same
frequencies as unlicensed frequencies may vary from one country to another. The
LoRaWAN specification defines a set of default frequencies (known as channels) for
each region of the globe and some associated mechanisms for its RX1 and RX2 receiving
windows that may vary from one region to another.

This module provides some default classes that represent the default channel plans
for different regions but also let the user free to create a new one based on his
needs. 

A channel plan is defined by a set of uplink and downlink channels, a special backup
downlink channel (called RX2), and a mechanism used to deduce the first receiving channel
characteristics. In our implementation, the channel plan also embeds a set of datarates
and the supported datarate for each channel. This deviates from the specification but
since the hardware we are using does not accept different datarates on a single channel,
we don't have any other choice than to assign a single datarate to each channel.

A simple channel plan is defined as follows:

``` python
class MyChannelPlan(ChannelPlan):

    def __init__(self):
        super().__init__(
            channels=[
                Uplink(1, 868100000, 0),
                Downlink(1, 868100000, 1),
            ],
            datarates=[
                DataRate(sf=7, bw=125000),
                DataRate(sf=12, bw=125000)
            ],
            rx2=Downlink(2, 868269000, 1)
        )
```

This channel plan defines a single uplink channel on 868.1 MHz using datarate 0 that is defined
as a spreading factor (sf) of 7 and a bandwidth of 125 kHz (first datarate in our `datarates` parameter).
A downlink channel is also declared using the same frequency than the uplink channel but a different
datarate (spreading factor of 12 and bandwidth of 125 kHz). Last but not least, a backup downlink channel
is defined using the 868.268 MHz frequency and datarate 1 (spreading factor of 12 and bandwidth of 125 kHz).

RX1 selection behavior
----------------------

When a LoRaWAN gateway successfully receives a frame on one of its uplink channels, it then needs to switch
to a downlink channel (if required) to send some data back to the device that sent the frame. This downlink
channel is usually chosen based on the uplink channel characteristicsm therefore the `ChannelPlan` class
provides a `get_rx1()` method that can be overriden to implement a specific RX1 selection behavior.

This method takes the uplink channel number in parameter allowing to perform some computations to determine
the downlink channel to use.

"""
from random import choice

from .exceptions import ChannelNotFound, InvalidDataRate

class Channel(object):
    """This class represents a LoRaWAN channel that is used in
    a frequency plan.
    """

    def __init__(self, number: int, frequency: int, data_rate: int=0):
        """Initialize the channel information.
        """
        self.__number = number
        self.__frequency = frequency
        self.__dr = data_rate


    def __repr__(self):
        return 'Channel(num=%d, freq=%d, DR%d)' % (
            self.__number,
            self.__frequency,
            self.__dr,
        )

    @property
    def number(self):
        return self.__number

    @property
    def frequency(self):
        """Retrieve the channel frequency.
        """
        return self.__frequency
        
    @property
    def data_rate(self):
        """Retrieve the supported bandwidth
        """
        return self.__dr
    
    @data_rate.setter
    def data_rate(self, value: int=0):
        self.__dr = value


class Downlink(Channel):
    '''Downlink channel
    '''
    def __init__(self, number: int, frequency: int, data_rate: int=0):
        super().__init__(number=number, frequency=frequency, data_rate=data_rate)


class Uplink(Channel):
    '''Uplink channel
    '''
    def __init__(self, number: int, frequency: int, data_rate: int=0):
        super().__init__(number=number, frequency=frequency, data_rate=data_rate)


class ChannelModParams(Channel):
    '''Channel modulation parameters.
    '''
    def __init__(self, frequency: int, data_rate: int=0, sf: int=7, bw: int=125000, number: int=0):
        super().__init__(number=number, frequency=frequency, data_rate=data_rate)
        self.__sf = sf
        self.__bw = bw

    def __repr__(self):
        return 'ChannelModParams(number=%d, freq=%d, DR%d, sf=%d, bw=%d)' % (
            self.number,
            self.frequency,
            self.data_rate,
            self.spreading_factor,
            self.bandwidth
        )

    def __eq__(self, other):
        '''Channels are equivalent if same frequency, spreading factor and bandwidth
        '''
        return (self.frequency == other.frequency) and \
            (self.spreading_factor == other.spreading_factor) and \
            (self.bandwidth == other.bandwidth)

    @property
    def spreading_factor(self):
        return self.__sf
    
    @property
    def bandwidth(self):
        return self.__bw



class DataRate(object):
    '''DataRate model
    '''

    def __init__(self, sf: int = 7, bw: int = 125000):
        self.__sf = sf
        self.__bw = bw

    @property
    def spreading_factor(self):
        return self.__sf
    
    @property
    def bandwidth(self):
        return self.__bw


class ChannelPlan(object):
    '''LoRa Frequency Plan
    '''

    def __init__(self, channels: [Channel], datarates: [DataRate], rx2: Channel = None):
        '''Loop on channel provided in args
        '''
        self.__uplink_channels = {}
        self.__downlink_channels = {}
        self.__datarates = datarates
        self.__rx2 = rx2

        for channel in channels:
            if isinstance(channel, Downlink):
                self.__downlink_channels[channel.number] = channel
            else: 
                self.__uplink_channels[channel.number] = channel

    def get_rx1(self, chan_number: int) -> ChannelModParams:
        '''Retrieve RX1 channel based on TX channel.

        :param chan_number: tx channel number
        :type chan_number: int
        :return: RX1 channel modulation parameters
        '''
        if chan_number in self.__downlink_channels:
            channel = self.__downlink_channels[chan_number]
            return ChannelModParams(
                channel.frequency,
                channel.data_rate,
                self.__datarates[channel.data_rate].spreading_factor,
                self.__datarates[channel.data_rate].bandwidth,
                number=channel.number
            )
        else:
            raise ChannelNotFound
    
    def get_rx2(self) -> ChannelModParams:
        """Retrieve the channel defined for RX2

        :return: RX2 channel
        """
        # Resolve spreading factor and bandwidth from datarate
        if self.__rx2.data_rate < len(self.__datarates):
            return ChannelModParams(
                self.__rx2.frequency,
                self.__rx2.data_rate,
                self.__datarates[self.__rx2.data_rate].spreading_factor,
                self.__datarates[self.__rx2.data_rate].bandwidth,
                number=self.__rx2.number
            )
        else:
            raise InvalidDataRate

    def has_uplink(self, chan_index):
        '''Determine if a channel number is in the uplink channel plan
        '''
        return (chan_index in self.__uplink_channels)

    def has_downlink(self, chan_index):
        '''Determine if a channel number is in the downlink channel plan
        '''
        return (chan_index in self.__uplink_channels)

    def get_uplink(self, chan_index: int):
        '''Retrieve an uplink channel given its index
        '''
        if chan_index in self.__uplink_channels:
            return self.__uplink_channels[chan_index]
        else:
            raise ChannelNotFound
        
    def get_downlink(self, chan_index: int):
        '''Retrieve a downlink channel given its index
        '''
        if chan_index in self.__downlink_channels:
            return self.__downlink_channels[chan_index]
        else:
            raise ChannelNotFound
    
    def channels(self):
        '''Channels iterator (uplink and downlink)
        '''
        for channel in self.__uplink_channels.values():
            yield ChannelModParams(
                channel.frequency,
                channel.data_rate,
                self.__datarates[channel.data_rate].spreading_factor,
                self.__datarates[channel.data_rate].bandwidth,
                channel.number                
            )
        for channel in self.__downlink_channels.values():
            yield ChannelModParams(
                channel.frequency,
                channel.data_rate,
                self.__datarates[channel.data_rate].spreading_factor,
                self.__datarates[channel.data_rate].bandwidth,
                channel.number                
            )

    def pick_channel(self) -> ChannelModParams:
        """Select a channel from our frequency plan, based on our criterias.
        """
        if len(self.__uplink_channels) > 0:
            chan_index = choice(list(self.__uplink_channels.keys()))
            channel = self.__uplink_channels[chan_index]
            if channel.data_rate < len(self.__datarates):
                return ChannelModParams(
                    channel.frequency,
                    channel.data_rate,
                    self.__datarates[channel.data_rate].spreading_factor,
                    self.__datarates[channel.data_rate].bandwidth,
                    channel.number
                )
            else:
                raise InvalidDataRate
        else:
            raise ChannelNotFound


class EU868(ChannelPlan):
    """LoRaWAN Europe 868MHz frequency plan
    """
    def __init__(self):
        super().__init__(
            channels = [
                # Uplink (1-8, DR5)
                Uplink(1, 868100000, 5),
                Uplink(2, 868300000, 5),
                Uplink(3, 868500000, 5),
                Uplink(4, 867100000, 5),
                Uplink(5, 867300000, 5),
                Uplink(6, 867500000, 5),
                Uplink(7, 867700000, 5),
                Uplink(8, 867900000, 5),
                Downlink(1, 868100000, 5),
                Downlink(2, 868300000, 5),
                Downlink(3, 868500000, 5),
                Downlink(4, 867100000, 5),
                Downlink(5, 867300000, 5),
                Downlink(6, 867500000, 5),
                Downlink(7, 867700000, 5),
                Downlink(8, 867900000, 5),
            ],

            datarates = [
                DataRate(12, 125000),
                DataRate(11, 125000),
                DataRate(10, 125000),
                DataRate(9, 125000),
                DataRate(8, 125000),
                DataRate(7, 125000),
            ],

            # Downlink RX2, DR0
            rx2=Downlink(10, 869525000, 0)
        )




from whad.common.monitors.pcap import PcapWriterMonitor
from whad.exceptions import ExternalToolNotFound
from whad.esb.connector import ESB
from whad.unifying.connector import Unifying
from tempfile import gettempdir, _get_candidate_names
from subprocess import Popen, DEVNULL
import whad

from shutil import which
from os import mkfifo, unlink
from os.path import dirname, realpath, exists
from time import sleep

class WiresharkMonitor(PcapWriterMonitor):
    """
    WiresharkMonitor.

    Runs a wireshark instance in background and monitor the traffic received and transmitted
    by the targeted connector. It is mainly a very basic wrapper that launches wireshark in background,
    creates a named fifo and populates it using underlying PcapWriterMonitor implementation.
    """

    @classmethod
    def get_dissector(cls, dissector_name):
        return realpath("{}/resources/wireshark/{}.lua".format(dirname(whad.__file__), dissector_name))

    @classmethod
    def get_dlt(cls, domain):
        dlt = None
        return dlt

    def __init__(self, monitor_reception=True, monitor_transmission=True):
        self._wireshark_process = None
        # Checks the presence of wireshark
        self._wireshark_path = which("wireshark")
        if self._wireshark_path is None:
            raise ExternalToolNotFound("wireshark")
        # We create a random name for our named pipe.
        self.fifo_name = gettempdir()+"/" + next(_get_candidate_names()) + ".pcap"
        mkfifo(self.fifo_name)

        self.dissector = None
        super().__init__(
                            pcap_file=self.fifo_name,
                            monitor_reception=monitor_reception,
                            monitor_transmission=monitor_transmission
        )

    def attach(self, connector):
        """Attach to connector
        """
        if connector.domain in ("esb", "unifying"):
            self.dissector = WiresharkMonitor.get_dissector("esb")
        return super().attach(connector)

    def setup(self):
        self._start_wireshark(self.fifo_name, self.dissector)
        super().setup()

    def _start_wireshark(self, fifo, dissector=None):
        if dissector is None:
            self._wireshark_process = Popen([self._wireshark_path, "-k", "-i", fifo], stderr=DEVNULL, stdout=DEVNULL)
        else:
            with open(dissector, "r") as f:
                conf_line = [line for line in f.readlines() if "Proto(" in line][0]
                dissector_name = conf_line.split("Proto(")[1].split(",")[0].replace("\"", "")
            self._wireshark_process = Popen([self._wireshark_path,"-X","lua_script:"+dissector,"-o","uat:user_dlts:\"User 1 (DLT=148)\",\""+dissector_name+"\",\"\",\"\",\"\",\"\"", "-k", "-i", fifo], stderr=DEVNULL, stdout=DEVNULL)

    def is_terminated(self) -> bool:
        """Check if wireshark process has terminated.

        :return: `True` if process has terminated, `False` otherwise
        :rtype: bool
        """
        return self._wireshark_process.poll() is not None

    def close(self):
        self._writer_lock.acquire()
        if hasattr(self, "_writer") and self._writer is not None:


            # Close writer
            try:
                self._writer.close()

            except BrokenPipeError:
                pass

            # Mark writer as not available anymore
            self._writer = None

        # Release lock on writer
        self._writer_lock.release()

        if self._wireshark_process is not None:
            self._wireshark_process.terminate()
        if exists(self.fifo_name):
            unlink(self.fifo_name)



"""
Logitech Unifying applicative layer.
"""
from whad.esb.stack.llm.constants import ESBRole
from whad.scapy.layers.unifying import Logitech_Unifying_Hdr, Logitech_Mouse_Payload, Logitech_Set_Keepalive_Payload, \
    Logitech_Keepalive_Payload, Logitech_Unencrypted_Keystroke_Payload, Logitech_Encrypted_Keystroke_Payload, \
    Logitech_Multimedia_Key_Payload, Logitech_Waked_Up_Payload, Logitech_Wake_Up_Payload
from whad.unifying.hid import LogitechUnifyingMouseMovementConverter, LogitechUnifyingKeystrokeConverter, \
    HIDCodeNotFound, InvalidHIDData
from whad.unifying.stack.constants import UnifyingRole, ClickType, MultimediaKey
from whad.unifying.crypto import LogitechUnifyingCryptoManager
from whad.unifying.exceptions import MissingEncryptedKeystrokePayload
from whad.exceptions import WhadDeviceDisconnected, WhadDeviceTimeout
from whad.common.stack import Layer, alias, source, state, LayerState
from time import sleep, time
from threading import Thread, Lock
from queue import Queue, Empty
from whad.esb.esbaddr import ESBAddress
from sys import _getframe as getframe

import logging
logger = logging.getLogger(__name__)

class UnifyingApplicativeLayerState(LayerState):
    def __init__(self):
        super().__init__()
        self.role = UnifyingRole.DONGLE
        self.locale = "fr"
        self.wait_wakeup = False
        self.transmit_timeouts = False
        self.aes_counter = 0
        self.synchronized = False
        self.check_timeouts = False
        self.current_timeout = None
        self.last_timestamp = None
        self.last_sync = None

@alias('app')
@state(UnifyingApplicativeLayerState)
class UnifyingApplicativeLayer(Layer):

    def configure(self, options={}):
        self.__timeout_thread = None
        self.__crypto_manager = None
        self.callbacks = {}
        self.__packets_queue = Queue(10)

    def dongle_callback(func):
        def run_callback(*args, **kwargs):
            callbacks = getattr(args[0], "callbacks")
            execute = True
            if func.__name__ in callbacks and callable(callbacks[func.__name__]):
                execute = callbacks[func.__name__](*args[1:], **kwargs)
            if execute is None or execute:
                func(*args, **kwargs)
        return run_callback

    @property
    def aes_counter(self):
        return self.state.aes_counter

    @aes_counter.setter
    def aes_counter(self, counter):
        self.state.aes_counter = counter

    @property
    def locale(self):
        return self.state.locale

    @locale.setter
    def locale(self, value):
        self.state.locale = value

    def _check_timeouts_thread(self):
        while self.state.check_timeouts:
            if self.state.current_timeout is not None and self.state.last_timestamp is not None:
                if (int(time()*1000) - self.state.last_timestamp) > self.state.current_timeout * 10:
                    self.state.current_timeout = None
                    self.state.last_timestamp = None
                    self.on_desynchronized()
                else:
                    sleep(0.1)
            sleep(0.1)

    def _start_timeout_thread(self):
        self._stop_timeout_thread()
        self.state.transmit_timeouts = True
        self.__timeout_thread = Thread(target=self._transmit_timeouts_thread, daemon=True)
        self.__timeout_thread.start()

    def _stop_timeout_thread(self):
        if self.__timeout_thread is not None:
            self.state.transmit_timeouts = False
            self.__timeout_thread.join()
            self.__timeout_thread = None


    def _start_check_timeout_thread(self):
        self._stop_check_timeout_thread()
        self.state.check_timeouts = True
        self.__timeout_thread = Thread(target=self._check_timeouts_thread, daemon=True)
        self.__timeout_thread.start()

    def _stop_check_timeout_thread(self):
        if self.__timeout_thread is not None:
            self.state.check_timeouts = False
            self.__timeout_thread.join()
            self.__timeout_thread = None

    def _transmit_timeouts_thread(self):
        while self.state.transmit_timeouts or not self.__packets_queue.empty():
            try:
                self.send_message(Logitech_Set_Keepalive_Payload(timeout=1250))
                packet = self.__packets_queue.get(timeout=0.01, block=False)
                self.send_message(packet)
                self.send_message(Logitech_Keepalive_Payload(timeout=1250))
            except Empty as empty_err:
                try:
                    self.send_message(Logitech_Keepalive_Payload(timeout=1250))
                    sleep(0.01)
                except WhadDeviceTimeout:
                    return
                except WhadDeviceDisconnected:
                    return
            except WhadDeviceTimeout:
                return
            except WhadDeviceDisconnected:
                return

    def send_message(self, message, waiting_ack=False):
        result = self.send('ll', Logitech_Unifying_Hdr()/message, tag='data')
        return result

    def prepare_message(self, message):
        self.__packets_queue.put(message)
        return True

    @property
    def role(self):
        return self.state.role

    @role.setter
    def role(self, role):
        self.state.role = role
        if self.state.role == UnifyingRole.DONGLE:
            self.get_layer('ll').role = ESBRole.PRX
        else:
            self.get_layer('ll').role = ESBRole.PTX

    @property
    def key(self):
        if self.__crypto_manager is not None:
            return self.__crypto_manager.key
        else:
            return None

    @key.setter
    def key(self, key):
        if self.__crypto_manager is None or self.__crypto_manager.key != key:
            if key is not None:
                self.__crypto_manager = LogitechUnifyingCryptoManager(key)
            else:
                self.__crypto_manager = None

    def enable_timeouts(self):
        if self.state.role == UnifyingRole.DONGLE:
            self._start_check_timeout_thread()
        else:
            self._start_timeout_thread()

    def lock_channel(self):
        if self.__timeout_thread is None:
            logger.info('locking channel.')
            self.enable_timeouts()

    def unlock_channel(self):
        if self.__timeout_thread is not None:
            logger.info('unlocking channel.')
            self._stop_timeout_thread()

    def move_mouse(self, x, y):
        self.lock_channel()
        try:
            answer = self.prepare_message(
                Logitech_Mouse_Payload(
                    movement=LogitechUnifyingMouseMovementConverter.get_hid_data_from_coordinates(x, y)
                )
            )
            return answer
        except InvalidHIDData:
            return False

    def click_mouse(self, type=ClickType.RIGHT):
        self.lock_channel()
        answer = self.prepare_message(
            Logitech_Mouse_Payload(
                button_mask=int(type)
            )
        )
        return answer

    def wheel_mouse(self, x, y):
        self.lock_channel()
        answer = self.prepare_message(
            Logitech_Mouse_Payload(
                button_mask=0,
                movement='',
                wheel_x=x,
                wheel_y=y
            )

        )
        return answer

    def unencrypted_keystroke(self, key, ctrl=False, alt=False, shift=False, gui=False):
        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                Logitech_Unencrypted_Keystroke_Payload(
                    hid_data=LogitechUnifyingKeystrokeConverter.get_hid_data_from_key(
                        key,
                        ctrl=ctrl,
                        alt=alt,
                        shift=shift,
                        gui=gui,
                        locale=self.state.locale
                    )
                )
            )

            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                Logitech_Unencrypted_Keystroke_Payload(
                    hid_data=b"\x00"*7
                )
            )

            return answer_press and answer_release

        except InvalidHIDData:
            return False

    def multimedia_keystroke(self,key):
        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                Logitech_Multimedia_Key_Payload(
                    hid_key_scan_code=b"x\00"+bytes([int(key)])+b"\x00"*2
                )
            )
            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                Logitech_Multimedia_Key_Payload(
                    hid_key_scan_code=b"\x00"*4
                )
            )

            return answer_press and answer_release

        except InvalidHIDData:
            return False


    def encrypted_keystroke(self, key, ctrl=False, alt=False, shift=False, gui=False, force_counter=None):
        if self.__crypto_manager is None:
            return False

        counter = self.state.aes_counter if force_counter is None else force_counter

        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                self.__crypto_manager.encrypt(
                    Logitech_Encrypted_Keystroke_Payload(
                        hid_data=LogitechUnifyingKeystrokeConverter.get_hid_data_from_key(
                            key,
                            ctrl=False,
                            alt=False,
                            shift=False,
                            gui=False,
                            locale=self.state.locale
                        ),
                        unknown=201,
                        aes_counter=counter
                    )
                )
            )
            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                self.__crypto_manager.encrypt(

                    Logitech_Encrypted_Keystroke_Payload(
                        hid_data=b"\x00"*7,
                        unknown=201,
                        aes_counter=counter+1
                    )
                )
            )

            if force_counter is None:
                self.aes_counter += 2

            return answer_press and answer_release
        except InvalidHIDData:
            return False

    #@dongle_callback
    @source('ll', 'synchronized')
    def on_synchronized(self, timestamp=None):
        if timestamp is None:
            timestamp = time()
        logger.info("Synchronized !")
        if self.state.role == UnifyingRole.DONGLE:
            self.enable_timeouts()
        self.state.synchronized = True

    #@dongle_callback
    @source('ll', 'desynchronized')
    def on_desynchronized(self, timestamp=None):
        #print("desync")
        if timestamp is None:
            timestamp = time()
        logger.info("Desynchronized.")
        self.state.synchronized = False

    def wait_synchronization(self):
        while not self.state.synchronized:
            sleep(0.01)
        return True

    def wait_wakeup(self):
        self.get_layer('ll').address = ESBAddress(self.get_layer('ll').address).base + ":00"
        self.state.wait_wakeup = True
        while self.state.wait_wakeup:
            sleep(0.01)
        return True

    #@dongle_callback
    @source('ll', 'data')
    def on_data(self, data):
        current_time = int(time() * 1000)
        self.state.last_timestamp = current_time

        if self.state.role == UnifyingRole.DONGLE:
            if not self.state.synchronized:
                self.on_synchronized()

            if self.state.wait_wakeup:
                if Logitech_Wake_Up_Payload in data:
                    # Weird checksum, force it
                    pkt = Logitech_Unifying_Hdr(dev_index = 0,checksum=0xAC)/Logitech_Waked_Up_Payload(wakeup_dev_index=data.dev_index)
                    self.send(pkt, tag='ack')
                elif hasattr(data,"dev_index"):
                    self.get_layer('ll').address = ESBAddress(self.get_layer('ll').address).base + ":{:02x}".format(data.dev_index)
                    self.state.wait_wakeup = False
                    self.on_wakeup(data.dev_index)

            if Logitech_Set_Keepalive_Payload in data:
                self.on_set_keepalive(data.timeout)

            if Logitech_Keepalive_Payload in data:
                self.on_keepalive(data.timeout)

            if Logitech_Mouse_Payload in data:
                self.on_mouse_payload(data)

            if Logitech_Multimedia_Key_Payload in data and data.hid_key_scan_code != b"\x00"*4:
                try:
                    self.on_multimedia_keystroke(MultimediaKey(data.hid_key_scan_code[0]))
                except ValueError:
                    pass
            if Logitech_Unencrypted_Keystroke_Payload in data:
                self.on_unencrypted_keystroke_payload(data)

            if Logitech_Encrypted_Keystroke_Payload in data:
                self.on_encrypted_keystroke_payload(data)

    @dongle_callback
    def on_multimedia_keystroke(self, key):
        logger.info("Multimedia keystroke (key="+str(key.name)+")")

    @dongle_callback
    def on_unencrypted_keystroke_payload(self, data):
        logger.info("Unencrypted Keystroke Payload (payload="+bytes(data).hex()+")")
        if data.hid_data != b"\x00"*7:
            try:
                key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(data.hid_data, locale=self.state.locale)
                self.on_unencrypted_keystroke(key)
            except (InvalidHIDData, HIDCodeNotFound):
                pass

    @dongle_callback
    def on_encrypted_keystroke_payload(self, data):
        logger.info("Encrypted Keystroke Payload (payload="+bytes(data).hex()+")")
        if self.__crypto_manager is not None:
            try:
                decrypted = self.__crypto_manager.decrypt(data)
                logger.info("\t-> Decrypted payload: ", bytes(decrypted).hex())
                if decrypted.hid_data != b"\x00"*7:
                    key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(decrypted.hid_data, locale=self.state.locale)
                    self.on_encrypted_keystroke(key)
            except (MissingEncryptedKeystrokePayload, InvalidHIDData, HIDCodeNotFound):
                pass

    @dongle_callback
    def on_encrypted_keystroke(self, key):
        logger.info("Encrypted keystroke (key="+str(key)+")")
        self.on_keystroke(key)

    @dongle_callback
    def on_unencrypted_keystroke(self, key):
        logger.info("Unencrypted keystroke (key="+str(key)+")")
        self.on_keystroke(key)

    @dongle_callback
    def on_keystroke(self, key):
        logger.info("Keystroke ("+str(key)+")")

    @dongle_callback
    def on_wakeup(self, dev_index):
        logger.info("Waked up by device (dev_index=0x%02x)", dev_index)

    @dongle_callback
    def on_set_keepalive(self, timeout):
        logger.info("Set keep alive (timeout="+str(timeout)+")")
        self.state.current_timeout = timeout

    @dongle_callback
    def on_keepalive(self, timeout):
        logger.info("Keep alive (timeout="+str(timeout)+")")
        self.state.current_timeout = timeout

    @dongle_callback
    def on_mouse_payload(self, data):
        try:
            logger.info("Mouse payload (payload="+bytes(data).hex()+")")
            converter = LogitechUnifyingMouseMovementConverter()
            x, y = converter.get_coordinates_from_hid_data(data.movement)
            if x != 0 or y != 0:
                self.on_move_mouse(x, y)

            button = ClickType(data.button_mask)
            if button != ClickType.NONE:
                self.on_click_mouse(button)

            if data.wheel_x != 0 or data.wheel_y != 0:
                self.on_wheel_mouse(data.wheel_x, data.wheel_y)
        except ValueError:
            pass

    @dongle_callback
    def on_wheel_mouse(self, x, y):
        logger.info("Mouse wheel (x="+str(x)+", y="+str(y)+")")

    @dongle_callback
    def on_move_mouse(self, x, y):
        logger.info("Mouse move (x="+str(x)+", y="+str(y)+")")

    @dongle_callback
    def on_click_mouse(self, type):
        logger.info("Mouse click (click="+str(type.name)+")")

    @source('ll', 'ack')
    def on_acknowledgement(self, ack):
        pass

    def __del__(self):
        self.stop()

    def stop(self):
        if self.__timeout_thread is not None:
            if self.state.role == UnifyingRole.DONGLE:
                self._stop_check_timeout_thread()
            else:
                self._stop_timeout_thread()

'''
class UnifyingApplicativeLayerManager:
    """
    This class handles the logic of Logitech Unifying protocol, used by Logitech mices and keyboards.
    The protocol is an application layer built over the Enhanced ShockBurst protocol, provided by Nordic SemiConductors.
    """
    def __init__(self, llm, role=UnifyingRole.DONGLE):
        self.__llm = llm
        self.__role = role
        self.__locale = "fr"
        self.__transmit_timeouts = False
        self.__timeout_thread = None
        self.__crypto_manager = None
        self.__aes_counter = 0
        self.__wait_wakeup = False
        self.__synchronized = False
        self.callbacks = {}
        self.__packets_queue = Queue(10)


    def dongle_callback(func):
        def run_callback(*args, **kwargs):
            callbacks = getattr(args[0], "callbacks")
            execute = True
            if func.__name__ in callbacks and callable(callbacks[func.__name__]):
                execute = callbacks[func.__name__](*args[1:], **kwargs)
            if execute is None or execute:
                func(*args, **kwargs)
        return run_callback

    @property
    def aes_counter(self):
        return self.__aes_counter

    @aes_counter.setter
    def aes_counter(self, counter):
        self.__aes_counter = counter

    @property
    def locale(self):
        return self.__locale

    @locale.setter
    def locale(self, value):
        self.__locale = value

    def _start_timeout_thread(self):
        self._stop_timeout_thread()
        self.__transmit_timeouts = True
        self.__timeout_thread = Thread(target=self._transmit_timeouts_thread)
        self.__timeout_thread.start()

    def _stop_timeout_thread(self):
        if self.__timeout_thread is not None:
            self.__transmit_timeouts = False
            self.__timeout_thread.join()
            self.__timeout_thread = None

    def _transmit_timeouts_thread(self):
        while self.__transmit_timeouts or not self.__packets_queue.empty():
            self.send_message(Logitech_Set_Keepalive_Payload(timeout=1250))
            try:
                packet = self.__packets_queue.get(timeout=0.01, block=False)
                self.send_message(packet)
                self.send_message(Logitech_Keepalive_Payload(timeout=1250))
            except Empty:
                self.send_message(Logitech_Keepalive_Payload(timeout=1250))
                sleep(0.01)

    def send_message(self, message, waiting_ack=False):
        result = self.__llm.send_data(Logitech_Unifying_Hdr()/message)
        return result

    def prepare_message(self, message):
        self.__packets_queue.put(message)
        return True

    def __del__(self):
        self.unlock_channel()

    @property
    def role(self):
        return self.__role

    @role.setter
    def role(self, role):
        self.__role = role
        if self.__role == UnifyingRole.DONGLE:
            self.__llm.role = ESBRole.PRX
        else:
            self.__llm.role = ESBRole.PTX

    @property
    def key(self):
        if self.__crypto_manager is not None:
            return self.__crypto_manager.key
        else:
            return None

    @key.setter
    def key(self, key):
        if self.__crypto_manager is None or self.__crypto_manager.key != key:
            if key is not None:
                self.__crypto_manager = LogitechUnifyingCryptoManager(key)
            else:
                self.__crypto_manager = None

    def enable_timeouts(self):
        if self.__role == UnifyingRole.DONGLE:
            raise RequiredImplementation("WaitingKeepAlives")
        else:
            self._start_timeout_thread()

    def lock_channel(self):
        if self.__timeout_thread is None:
            print("[i] locking channel.")
            self.enable_timeouts()

    def unlock_channel(self):
        if self.__timeout_thread is not None:
            print("[i] unlocking channel.")
            self._stop_timeout_thread()

    def move_mouse(self, x, y):
        self.lock_channel()
        try:
            answer = self.prepare_message(
                Logitech_Mouse_Payload(
                    movement=LogitechUnifyingMouseMovementConverter.get_hid_data_from_coordinates(x, y)
                )
            )
            return answer
        except InvalidHIDData:
            return False

    def click_mouse(self, type=ClickType.RIGHT):
        self.lock_channel()
        answer = self.prepare_message(
            Logitech_Mouse_Payload(
                button_mask=int(type)
            )
        )
        return answer

    def wheel_mouse(self, x, y):
        self.lock_channel()
        answer = self.prepare_message(
            Logitech_Mouse_Payload(
                button_mask=0,
                movement='',
                wheel_x=x,
                wheel_y=y
            )

        )
        return answer

    def unencrypted_keystroke(self, key, ctrl=False, alt=False, shift=False, gui=False):
        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                Logitech_Unencrypted_Keystroke_Payload(
                    hid_data=LogitechUnifyingKeystrokeConverter.get_hid_data_from_key(
                        key,
                        ctrl=False,
                        alt=False,
                        shift=False,
                        gui=False,
                        locale=self.__locale
                    )
                )
            )
            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                Logitech_Unencrypted_Keystroke_Payload(
                    hid_data=b"\x00"*7
                )
            )

            return answer_press and answer_release

        except InvalidHIDData:
            return False

    def multimedia_keystroke(self,key):
        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                Logitech_Multimedia_Key_Payload(
                    hid_key_scan_code=b"x\00"+bytes([int(key)])+b"\x00"*2
                )
            )
            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                Logitech_Multimedia_Key_Payload(
                    hid_key_scan_code=b"\x00"*4
                )
            )

            return answer_press and answer_release

        except InvalidHIDData:
            return False


    def encrypted_keystroke(self, key, ctrl=False, alt=False, shift=False, gui=False, force_counter=None):
        if self.__crypto_manager is None:
            return False

        counter = self.aes_counter if force_counter is None else force_counter

        self.lock_channel()
        try:
            answer_press = self.prepare_message(
                self.__crypto_manager.encrypt(
                    Logitech_Encrypted_Keystroke_Payload(
                        hid_data=LogitechUnifyingKeystrokeConverter.get_hid_data_from_key(
                            key,
                            ctrl=False,
                            alt=False,
                            shift=False,
                            gui=False,
                            locale=self.__locale
                        ),
                        unknown=201,
                        aes_counter=counter
                    )
                )
            )
            keep_alive = self.prepare_message(
                Logitech_Keepalive_Payload(timeout=1250)
            )
            answer_release = self.prepare_message(
                self.__crypto_manager.encrypt(

                    Logitech_Encrypted_Keystroke_Payload(
                        hid_data=b"\x00"*7,
                        unknown=201,
                        aes_counter=counter+1
                    )
                )
            )

            if force_counter is None:
                self.aes_counter += 2

            return answer_press and answer_release
        except InvalidHIDData:
            return False

    @dongle_callback
    def on_synchronized(self):
        print("[i] Synchronized !")
        self.__synchronized = True

    @dongle_callback
    def on_desynchronized(self):
        print("[i] Desynchronized.")
        self.__synchronized = False

    def wait_synchronization(self):
        while not self.__synchronized:
            sleep(0.01)
        return True

    def wait_wakeup(self):
        self.__llm.address = ESBAddress(self.__llm.address).base + ":00"
        self.__wait_wakeup = True
        while not self.__wait_wakeup:
            sleep(0.01)
        return True

    @dongle_callback
    def on_data(self, data):
        if self.__role == UnifyingRole.DONGLE:
            if not self.__synchronized:
                self.on_synchronized()

            if self.__wait_wakeup:
                if Logitech_Wake_Up_Payload in data:
                    # Weird checksum, force it
                    pkt = Logitech_Unifying_Hdr(dev_index = 0,checksum=0xAC)/Logitech_Waked_Up_Payload(wakeup_dev_index=data.dev_index)
                    self.__llm.prepare_acknowledgment(
                        pkt
                    )
                elif hasattr(data,"dev_index"):
                    self.__llm.address = ESBAddress(self.__llm.address).base + ":{:02x}".format(data.dev_index)
                    self.__wait_wakeup = False
                    self.on_wakeup(data.dev_index)

            if Logitech_Set_Keepalive_Payload in data:
                self.on_set_keepalive(data.timeout)

            if Logitech_Keepalive_Payload in data:
                self.on_keepalive(data.timeout)

            if Logitech_Mouse_Payload in data:
                self.on_mouse_payload(data)

            if Logitech_Multimedia_Key_Payload in data and data.hid_key_scan_code != b"\x00"*4:
                try:
                    self.on_multimedia_keystroke(MultimediaKey(data.hid_key_scan_code[0]))
                except ValueError:
                    pass
            if Logitech_Unencrypted_Keystroke_Payload in data:
                self.on_unencrypted_keystroke_payload(data)

            if Logitech_Encrypted_Keystroke_Payload in data:
                self.on_encrypted_keystroke_payload(data)

    @dongle_callback
    def on_multimedia_keystroke(self, key):
        print("[i] Multimedia keystroke (key="+str(key.name)+")")

    @dongle_callback
    def on_unencrypted_keystroke_payload(self, data):
        print("[i] Unencrypted Keystroke Payload (payload="+bytes(data).hex()+")")
        if data.hid_data != b"\x00"*7:
            try:
                key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(data.hid_data, locale=self.__locale)
                self.on_unencrypted_keystroke(key)
            except (InvalidHIDData, HIDCodeNotFound):
                pass

    @dongle_callback
    def on_encrypted_keystroke_payload(self, data):
        print("[i] Encrypted Keystroke Payload (payload="+bytes(data).hex()+")")
        if self.__crypto_manager is not None:
            try:
                decrypted = self.__crypto_manager.decrypt(data)
                print("\t-> Decrypted payload: ", bytes(decrypted).hex())
                if decrypted.hid_data != b"\x00"*7:
                    key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(decrypted.hid_data, locale=self.__locale)
                    self.on_encrypted_keystroke(key)
            except (MissingEncryptedKeystrokePayload, InvalidHIDData, HIDCodeNotFound):
                pass

    @dongle_callback
    def on_encrypted_keystroke(self, key):
        print("[i] Encrypted keystroke (key="+str(key)+")")
        self.on_keystroke(key)

    @dongle_callback
    def on_unencrypted_keystroke(self, key):
        print("[i] Unencrypted keystroke (key="+str(key)+")")
        self.on_keystroke(key)

    @dongle_callback
    def on_keystroke(self, key):
        print("[i] Keystroke ("+str(key)+")")

    @dongle_callback
    def on_wakeup(self, dev_index):
        print("[i] Waked up by device (dev_index=0x{:02x})".format(dev_index))

    @dongle_callback
    def on_set_keepalive(self, timeout):
        print("[i] Set keep alive (timeout="+str(timeout)+")")

    @dongle_callback
    def on_keepalive(self, timeout):
        print("[i] Keep alive (timeout="+str(timeout)+")")

    @dongle_callback
    def on_mouse_payload(self, data):
        print("[i] Mouse payload (payload="+bytes(data).hex()+")")
        converter = LogitechUnifyingMouseMovementConverter()
        x, y = converter.get_coordinates_from_hid_data(data.movement)
        if x != 0 or y != 0:
            self.on_move_mouse(x, y)

        button = ClickType(data.button_mask)
        if button != ClickType.NONE:
            self.on_click_mouse(button)

        if data.wheel_x != 0 or data.wheel_y != 0:
            self.on_wheel_mouse(data.wheel_x, data.wheel_y)

    @dongle_callback
    def on_wheel_mouse(self, x, y):
        print("[i] Mouse wheel (x="+str(x)+", y="+str(y)+")")

    @dongle_callback
    def on_move_mouse(self, x, y):
        print("[i] Mouse move (x="+str(x)+", y="+str(y)+")")

    @dongle_callback
    def on_click_mouse(self, type):
        print("[i] Mouse click (click="+str(type.name)+")")

    def on_acknowledgement(self, ack):
        pass
'''



"""Ubertooth adaptation layer.
"""
import logging

from struct import unpack, pack
from time import sleep
from usb.core import find, USBError
from usb.util import get_string

from scapy.layers.bluetooth4LE import BTLE, BTLE_ADV, BTLE_CONNECT_REQ, BTLE_DATA
from scapy.compat import raw

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied
from whad.device import VirtualDevice
from whad.hub.discovery import Domain, Capability
from whad.ble.utils.phy import channel_to_frequency, frequency_to_channel, crc,\
    FieldsSize, is_access_address_valid
from whad.hub.ble import Direction, ChannelMap
from whad.hub.generic.cmdresult import CommandResult
from whad.hub.ble import Commands
from whad.scapy.layers.ubertooth import Ubertooth_Hdr,UBERTOOTH_PACKET_TYPES, \
    BTLE_Promiscuous_Access_Address, BTLE_Promiscuous_CRCInit, BTLE_Promiscuous_Hop_Interval, \
    BTLE_Promiscuous_Hop_Increment

from .constants import UbertoothId, \
    UbertoothTransfers, UbertoothModulations, UbertoothCommands, \
    UbertoothInternalState, UbertoothModes, UbertoothJammingModes

logger = logging.getLogger(__name__)

# Helpers functions
def get_ubertooth(index: int = 0, serial: str = None):
    """
    Returns an ubertooth USB object based on index or serial number.
    """
    devices = list(find(idVendor=UbertoothId.UBERTOOTH_ID_VENDOR,
                        idProduct=UbertoothId.UBERTOOTH_ID_PRODUCT,
                        find_all=True))
    if serial is not None:
        for device in devices:
            if serial.lower() == get_string(device, device.iSerialNumber):
                return (devices.index(device), device)
        # No device found with the corresponding serial, return None
        return None

    try:
        return (index, devices[index])
    except IndexError:
        return None

class Ubertooth(VirtualDevice):
    """Ubertooth virtual device implementation.
    """

    INTERFACE_NAME = "ubertooth"

    @classmethod
    def list(cls):
        """
        Returns a list of available Ubertooth devices.
        """
        available_devices = []
        try:
            for ubertooth in find(idVendor=UbertoothId.UBERTOOTH_ID_VENDOR,
                                  idProduct=UbertoothId.UBERTOOTH_ID_PRODUCT,
                                  find_all=True):
                available_devices.append(
                    Ubertooth(serial=get_string(ubertooth, ubertooth.iSerialNumber)
                ))
        except ValueError:
            logger.warning("Cannot access Ubertooth, root privileges may be required.")
        return available_devices

    @property
    def identifier(self):
        """
        Returns the identifier of the current device (e.g., serial number).
        """
        return get_string(self.__ubertooth, self.__ubertooth.iSerialNumber)


    def __init__(self, index=0, serial=None):
        """
        Create device connection
        """
        device = get_ubertooth(index,serial)
        if device is None:
            raise WhadDeviceNotFound()

        self.__opened = False
        self.__address_filter = b"\xFF\xFF\xFF\xFF\xFF\xFF"
        self.__access_address = 0x8e89bed6
        self.__hop_interval = 0
        self.__hop_increment = 0
        self.__channel_map = 0
        self.__crc_init = 0x555555
        self.__channel = 37
        self.__show_advertisements = True
        self.__show_empty_packets = False
        self.__internal_state = UbertoothInternalState.NONE
        self.__index, self.__ubertooth = device
        super().__init__()

    def open(self):
        try:
            self.__ubertooth.set_configuration()
        except USBError as err:
            # Access denied ?
            if err.errno == 13:
                raise WhadDeviceAccessDenied("ubertooth") from err
            
            # Device not ready
            raise WhadDeviceNotReady() from err

        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

        self.__opened = True
        # Ask parent class to run a background I/O thread
        super().open()

    def write(self, data: bytes):
        """handle write operation.

        :param data: Data to write
        :type data: bytes
        :raises: WhadDeviceNotReady
        """
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        """Process incoming data.
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        if self.__internal_state != UbertoothInternalState.NONE:
            data = self._ubertooth_ctrl_transfer_in(UbertoothCommands.UBERTOOTH_POLL,512)
            if len(data) > 0:
                packet = Ubertooth_Hdr(data)
                self.__channel = frequency_to_channel(2402+packet.channel)
                if UBERTOOTH_PACKET_TYPES[packet.packet_type] == "LE_PACKET":
                    timestamp = round(packet.clk_100ns/10) # convert into us timestamp
                    rssi = packet.rssi_min - 54
                    if self.__internal_state == UbertoothInternalState.ACCESS_ADDRESS_SNIFFING:
                        access_address = packet[BTLE:].access_addr
                        if is_access_address_valid(access_address):
                            self._send_whad_ble_aa_disc(access_address, timestamp, rssi)
                    else:
                        btle_packet = packet[BTLE:]
                        if self._filter(btle_packet):
                            self._send_whad_ble_raw_pdu(btle_packet, timestamp, rssi)
                        if (
                                self.__internal_state == UbertoothInternalState.NEW_CONNECTION_SNIFFING and
                                BTLE_CONNECT_REQ in btle_packet
                        ):
                            self.__access_address = btle_packet.AA
                            self.__channel_map = btle_packet.chM
                            self.__crc_init = btle_packet.crc_init
                            self.__hop_interval = btle_packet.interval
                            self.__hop_increment = btle_packet.hop
                            self._send_whad_ble_synchronized()
                elif BTLE_Promiscuous_Access_Address in packet:
                    self.__crc_init = 0
                    self.__channel_map = 0
                    self.__hop_interval = 0
                    self.__hop_increment = 0
                    self.__access_address = packet.access_address
                    self._send_whad_ble_synchronized()
                elif BTLE_Promiscuous_CRCInit in packet:
                    self.__crc_init = packet.crc_init
                    self.__channel_map = 0
                    self.__hop_interval = 0
                    self.__hop_increment = 0
                    self._send_whad_ble_synchronized()
                elif BTLE_Promiscuous_Hop_Interval in packet:
                    self.__channel_map = 0x1fffffffff
                    self.__hop_interval = packet.hop_interval
                    self.__hop_increment = 0
                    self._send_whad_ble_synchronized()
                elif BTLE_Promiscuous_Hop_Interval in packet:
                    self.__hop_increment = packet.hop_increment
                    self._send_whad_ble_synchronized()

                else:
                    print(data.hex())
                    packet.show()

    def reset(self):
        self.__internal_state = UbertoothInternalState.NONE
        self.__ubertooth.reset()

        self._set_jam_mode(UbertoothJammingModes.JAM_NONE)
        self._set_modulation(UbertoothModulations.MOD_BT_LOW_ENERGY)
        self._reset_clock()

    def close(self):
        self._soft_reset()
        super().close()

    # Virtual device whad message builder
    def _send_whad_ble_raw_pdu(self, packet, timestamp=None, rssi=None):

        access_address = packet.access_addr
        pdu = raw(packet)[FieldsSize.ACCESS_ADDRESS_SIZE:-FieldsSize.CRC_SIZE]
        sniffed_crc = raw(packet)[-FieldsSize.CRC_SIZE:]
        calculated_crc = crc(pdu,init=self.__crc_init)
        is_crc_valid = calculated_crc == sniffed_crc

        # Create a RawPduReceived message
        msg = self.hub.ble.create_raw_pdu_received(
            Direction.UNKNOWN,
            pdu,
            access_address,
            0,
            crc_validity=is_crc_valid,
            crc=packet.crc,
            channel=self.__channel
        )

        # Set RSSI and timestamp if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message
        self._send_whad_message(msg)


    def _send_whad_ble_synchronized(self):
        # Create a Synchronized message
        msg = self.hub.ble.create_synchronized(
            self.__access_address,
            self.__hop_interval,
            self.__hop_increment,
            ChannelMap.from_int(self.__channel_map),
            self.__crc_init
        )

        # Send message
        self._send_whad_message(msg)


    def _send_whad_ble_aa_disc(self, access_address, timestamp, rssi):
        # Create an AccessAddressDiscovered
        msg = self.hub.ble.create_access_address_discovered(
            access_address,
            rssi,
            timestamp
        )

        # Send message
        self._send_whad_message(msg)

    # Virtual device whad message callbacks
    def _on_whad_ble_stop(self, message):
        self._stop()
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_conn(self, message):
        self.__access_address = message.access_address
        self._set_access_address(self.__access_address)
        self.__internal_state = UbertoothInternalState.EXISTING_CONNECTION_SNIFFING
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_adv(self, message):
        channel = message.channel
        # Address filtering is performed in python, ubertooth doesn't support
        # it natively for adv only mode
        self.__address_filter = message.bd_address

        self.__access_address = 0x8e89bed6
        self.__crc_init = 0x555555
        self.__show_advertisements = True

        if self._set_channel(channel):
            self.__internal_state = UbertoothInternalState.ADVERTISEMENT_SNIFFING
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_ble_sniff_connreq(self, message):
        channel = message.channel
        self.__address_filter = b"\xFF\xFF\xFF\xFF\xFF\xFF"
        self.__show_empty_packets = message.show_empty_packets
        self.__show_advertisements = message.show_advertisements
        if self._set_channel(channel):
            self.__internal_state = UbertoothInternalState.NEW_CONNECTION_SNIFFING
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_ble_sniff_aa(self, message):
        self.__show_empty_packets = False
        self.__show_advertisements = False
        self.__internal_state = UbertoothInternalState.ACCESS_ADDRESS_SNIFFING
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_start(self, message):
        if self.__internal_state == UbertoothInternalState.ADVERTISEMENT_SNIFFING:
            self._enable_advertisements_sniffing()
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == UbertoothInternalState.NEW_CONNECTION_SNIFFING:
            self._enable_connection_sniffing()
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state in (UbertoothInternalState.ACCESS_ADDRESS_SNIFFING,
                                       UbertoothInternalState.EXISTING_CONNECTION_SNIFFING):
            self._enable_promiscuous_mode()
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    # Software implementation features
    def _filter(self, packet):
        if BTLE_DATA in packet and packet.len == 0 and not self.__show_empty_packets:
            return False
        
        if BTLE_ADV in packet and not self.__show_advertisements:
            return False
        
        # No filtering if address is FF:FF:FF:FF:FF:FF
        if self.__address_filter == b"\xFF\xFF\xFF\xFF\xFF\xFF":
            return True
        
        # Ensure packet has the expected advertiser address.
        if hasattr(packet, "AdvA"):
            return packet.AdvA == self.__address_filter

        # Packet must be discarded.
        return False

    # Ubertooth low level communication primitives
    def _ubertooth_ctrl_transfer_in(self, request, size, timeout=100):
        try:
            received_data = self.__ubertooth.ctrl_transfer(UbertoothTransfers.CTRL_IN,
                                                           request, 0, 0, size,
                                                           timeout=timeout)
            received_data = received_data.tobytes()
            if len(received_data) == 1:
                received_data = b""
        except USBError:
            received_data = b""

        return received_data

    def _ubertooth_ctrl_transfer_out(self, request, value=0, data=None, timeout=100):
        try:
            self.__ubertooth.ctrl_transfer(
                UbertoothTransfers.CTRL_OUT,
                request,
                value,
                0,
                data,
                timeout=timeout
            )
            return True
        except USBError:
            return False

    # Discovery related functions
    def _get_capabilities(self):
        capabilities = {
            Domain.BtLE : (
                (Capability.Sniff | Capability.Jam),
                [
                    Commands.SniffAdv, Commands.SniffConnReq,
                    Commands.SniffAccessAddress, Commands.SniffActiveConn,
                    Commands.Start, Commands.Stop
                ]
            )
        }
        return capabilities

    def _get_serial_number(self):
        serial_number = self._ubertooth_ctrl_transfer_in(
            UbertoothCommands.UBERTOOTH_GET_SERIAL, 17
        )
        return serial_number

    def _get_url(self):
        url = "https://github.com/greatscottgadgets/ubertooth"
        return url.encode("utf-8")

    def _get_manufacturer(self):
        return get_string(self.__ubertooth, self.__ubertooth.iManufacturer).encode("utf-8")

    def _get_firmware_version(self):
        firmware_version = self._ubertooth_ctrl_transfer_in(
            UbertoothCommands.UBERTOOTH_GET_REV_NUM,20
        )[2:].decode("utf-8")
        if "git" in firmware_version:
            try:
                major, minor, revision = 1, 7, int(firmware_version.split("-")[1], 16)
            except ValueError:
                major, minor, revision = 1, 7, 0
        else:
            major, minor, revision = firmware_version.split("-")
            major, minor, revision = int(major), int(minor), int(revision.replace("R",""))
        return (major, minor, revision)

    # Ubertooth commands
    def _set_modulation(self, modulation=UbertoothModulations.MOD_BT_LOW_ENERGY):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_SET_MOD, modulation)

    def _stop(self):
        self.__internal_state = UbertoothInternalState.NONE
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_STOP)

    def _set_jam_mode(self, mode=UbertoothJammingModes.JAM_NONE):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_JAM_MODE, mode)

    def _reset_clock(self):
        self._ubertooth_ctrl_transfer_out(
            UbertoothCommands.UBERTOOTH_SET_CLOCK,
            data=b"\x00\x00\x00\x00\x00\x00"
        )

    def _set_crc_checking(self, enable=True):
        self._ubertooth_ctrl_transfer_out(
            UbertoothCommands.UBERTOOTH_SET_CRC_VERIFY,
            int(enable)
        )

    def _set_target(self, address=b"\x00\x00\x00\x00\x00\x00"):
        data = address + b"\x30"
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_BTLE_SET_TARGET, data=data)

    def _set_channel(self, channel: int = 37) -> bool:
        """Set channel.

        :param channel: Channel number to use
        :type channel: int
        :return: `True` on success, `False` otherwise
        :rtype: bool
        """
        frequency = channel_to_frequency(channel)
        if frequency is not None:
            self.__channel = channel
            self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_SET_CHANNEL, frequency)
            return True
        
        # Error while setting channel.
        return False

    def _get_channel(self):
        frequency = unpack('H', self._ubertooth_ctrl_transfer_in(
            UbertoothCommands.UBERTOOTH_GET_CHANNEL, 2)
        )[0]
        return frequency_to_channel(frequency)

    def _set_access_address(self, access_address):
        data = pack("<I", access_address)
        self._ubertooth_ctrl_transfer_out(
            UbertoothCommands.UBERTOOTH_SET_ACCESS_ADDRESS, data=data
        )

    def _get_access_address(self):
        access_address = unpack("<I",self._ubertooth_ctrl_transfer_in(
            UbertoothCommands.UBERTOOTH_GET_ACCESS_ADDRESS, 4)
        )[0]
        return access_address

    def _enable_promiscuous_mode(self):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_BTLE_PROMISC)

    def _enable_advertisements_sniffing(self):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_BTLE_SNIFFING, 0)

    def _enable_connection_sniffing(self):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_BTLE_SNIFFING, 2)

    def _soft_reset(self):
        self._ubertooth_ctrl_transfer_out(UbertoothCommands.UBERTOOTH_RESET)
